import json
import ujson
import logging
import inspect
import typing
import pydantic

from functools import lru_cache
from pydantic.fields import FieldInfo
from typing import Annotated, Callable, List, Tuple, Union  # noqa: UP035

import dspy
from dsp.adapters import passages2text
from dspy.primitives.prediction import Prediction
from dspy.signatures.signature import ensure_signature, make_signature

@lru_cache(maxsize=None)
def warn_once(msg: str):
    logging.warning(msg)


def predictor(*args: tuple, **kwargs) -> Callable[..., dspy.Module]:
    def _predictor(func) -> dspy.Module:
        """Decorator that creates a predictor module based on the provided function."""
        signature = _func_to_signature(func)
        *_, output_key = signature.output_fields.keys()
        return _StripOutput(TypedPredictor(signature, **kwargs), output_key)

    # if we have only a single callable argument, the decorator was invoked with no key word arguments
    #  so we just return the wrapped function
    if len(args) == 1 and callable(args[0]) and len(kwargs) == 0:
        return _predictor(args[0])
    return _predictor


def cot(*args: tuple, **kwargs) -> Callable[..., dspy.Module]:
    def _cot(func) -> dspy.Module:
        """Decorator that creates a chain of thought module based on the provided function."""
        signature = _func_to_signature(func)
        *_, output_key = signature.output_fields.keys()
        return _StripOutput(TypedChainOfThought(signature, **kwargs), output_key)

    # if we have only a single callable argument, the decorator was invoked with no key word arguments
    #  so we just return the wrapped function
    if len(args) == 1 and callable(args[0]) and len(kwargs) == 0:
        return _cot(args[0])
    return _cot


class _StripOutput(dspy.Module):
    def __init__(self, predictor, output_key):
        super().__init__()
        self.predictor = predictor
        self.output_key = output_key

    def copy(self):
        return _StripOutput(self.predictor.copy(), self.output_key)

    def forward(self, **kwargs):
        prediction = self.predictor(**kwargs)
        return prediction[self.output_key]


class FunctionalModule(dspy.Module):
    """To use the @cot and @predictor decorators, your module needs to inherit form this class."""

    def __init__(self):
        super().__init__()
        for name in dir(self):
            attr = getattr(self, name)
            if isinstance(attr, dspy.Module):
                self.__dict__[name] = attr.copy()


def TypedChainOfThought(signature, instructions=None, reasoning=None, *, max_retries=3) -> dspy.Module:  # noqa: N802
    """Just like TypedPredictor, but adds a ChainOfThought OutputField."""
    signature = ensure_signature(signature, instructions)
    output_keys = ", ".join(signature.output_fields.keys())

    default_rationale = dspy.OutputField(
        prefix="Reasoning: Let's think step by step in order to",
        desc="${produce the " + output_keys + "}. We ...",
    )
    reasoning = reasoning or default_rationale

    return TypedPredictor(
        signature.prepend(
            "reasoning",
            reasoning,
        ),
        max_retries=max_retries,
    )


class TypedPredictor(dspy.Module):
    def __init__(self, signature, instructions=None, *, max_retries=3, wrap_json=False, explain_errors=False):
        """Like dspy.Predict, but enforces type annotations in the signature.

        Args:
            signature: The signature of the module. Can use type annotations.
            instructions: A description of what the model should do.
            max_retries: The number of times to retry the prediction if the output is invalid.
            wrap_json: If True, json objects in the input will be wrapped in ```json ... ```
            explain_errors: If True, the model will try to explain the errors it encounters.
        """
        super().__init__()

        # Warn: deprecation warning.
        warn_once(
                "\t*** Since DSPy 2.5.16+, TypedPredictors are now deprecated, underperform, and are about to be removed! ***\n"
                "Please use standard predictors, e.g. dspy.Predict and dspy.ChainOfThought.\n"
                "They now support type annotations and other features of TypedPredictors and "
                "tend to work much better out of the box.\n"
                "Please let us know if you face any issues: https://github.com/stanfordnlp/dspy/issues"
            )

        signature = ensure_signature(signature, instructions)
        self.predictor = dspy.Predict(signature, _parse_values=False)
        self.max_retries = max_retries
        self.wrap_json = wrap_json
        self.explain_errors = explain_errors

    @property
    def signature(self) -> dspy.Signature:
        return self.predictor.signature

    @signature.setter
    def signature(self, value: dspy.Signature):
        self.predictor.signature = value

    def copy(self) -> "TypedPredictor":
        return TypedPredictor(
            self.signature,
            max_retries=self.max_retries,
            wrap_json=self.wrap_json,
            explain_errors=self.explain_errors,
        )

    def __repr__(self):
        """Return a string representation of the TypedPredictor object."""
        return f"TypedPredictor({self.signature})"

    def _make_example(self, field) -> str:
        # Note: DSPy will cache this call so we only pay the first time TypedPredictor is called.
        if hasattr(field, "model_json_schema"):
            pass
        schema = field.json_schema_extra["schema"]
        parser = field.json_schema_extra["parser"]
        if self.wrap_json:
            schema = "```json\n" + schema + "\n```\n"
        json_object = dspy.Predict(
            make_signature(
                "json_schema -> json_object",
                "Make a very succinct json object that validates with the following schema",
            ),
            _parse_values=False,
        )(json_schema=schema).json_object
        # We use the parser to make sure the json object is valid.
        try:
            parser(_unwrap_json(json_object, parser))
        except (pydantic.ValidationError, ValueError):
            return ""  # Unable to make an example
        return json_object
        # TODO: Another fun idea is to only (but automatically) do this if the output fails.
        # We could also have a more general "suggest solution" prompt that tries to fix the output
        # More directly.
        # TODO: Instead of using a language model to create the example, we can also just use a
        # library like https://pypi.org/project/polyfactory/ that's made exactly to do this.

    def _format_error(
        self,
        error: Exception,
        task_description: Union[str, FieldInfo],
        model_output: str,
        lm_explain: bool,
    ) -> str:
        if isinstance(error, pydantic.ValidationError):
            errors = []
            for e in error.errors():
                fields = ", ".join(map(str, e["loc"]))
                errors.append(f"{e['msg']}: {fields} (error type: {e['type']})")
            error_text = "; ".join(errors)
        else:
            error_text = repr(error)

        if self.explain_errors and lm_explain:
            if isinstance(task_description, FieldInfo):
                args = task_description.json_schema_extra
                task_description = args["prefix"] + " " + args["desc"]
            return (
                error_text
                + "\n"
                + self._make_explanation(
                    task_description=task_description,
                    model_output=model_output,
                    error=error_text,
                )
            )

        return error_text

    def _make_explanation(self, task_description: str, model_output: str, error: str) -> str:
        class Signature(dspy.Signature):
            """I gave my language model a task, but it failed.

            Figure out what went wrong, and write instructions to help it avoid the error next time.
            """

            task_description: str = dspy.InputField(desc="What I asked the model to do")
            language_model_output: str = dspy.InputField(desc="The output of the model")
            error: str = dspy.InputField(desc="The validation error triggered by the models output")
            explanation: str = dspy.OutputField(desc="Explain what the model did wrong")
            advice: str = dspy.OutputField(
                desc="Instructions for the model to do better next time. A single paragraph.",
            )

        # TODO: We could also try repair the output here. For example, if the output is a float, but the
        # model returned a "float + explanation", the repair could be to remove the explanation.

        return dspy.Predict(Signature)(
            task_description=task_description,
            language_model_output=model_output,
            error=error,
            _parse_values=False,
        ).advice

    def _prepare_signature(self) -> dspy.Signature:
        """Add formats and parsers to the signature fields, based on the type annotations of the fields."""
        signature = self.signature
        for name, field in self.signature.fields.items():
            is_output = field.json_schema_extra["__dspy_field_type"] == "output"
            type_ = field.annotation
            if is_output:
                if type_ is bool:

                    def parse(x):
                        x = x.strip().lower()
                        if x not in ("true", "false"):
                            raise ValueError("Respond with true or false")
                        return x == "true"

                    signature = signature.with_updated_fields(
                        name,
                        desc=field.json_schema_extra.get("desc", "")
                        + (" (Respond with true or false)" if type_ != str else ""),
                        format=lambda x: x if isinstance(x, str) else str(x),
                        parser=parse,
                    )
                elif type_ in (str, int, float):
                    signature = signature.with_updated_fields(
                        name,
                        desc=field.json_schema_extra.get("desc", "")
                        + (f" (Respond with a single {type_.__name__} value)" if type_ != str else ""),
                        format=lambda x: x if isinstance(x, str) else str(x),
                        parser=type_,
                    )
                else:
                    # Anything else we wrap in a pydantic object
                    if (
                        inspect.isclass(type_)
                        and typing.get_origin(type_) not in (list, tuple)  # To support Python 3.9
                        and issubclass(type_, pydantic.BaseModel)
                    ):
                        def to_json(x):
                            return x.model_dump_json()
                        def from_json(x, type_=type_):
                            return type_.model_validate_json(x)
                        schema = json.dumps(type_.model_json_schema())
                    else:
                        adapter = pydantic.TypeAdapter(type_)
                        def to_json(x):
                            return adapter.serializer.to_json(x)
                        def from_json(x, type_=adapter):
                            return type_.validate_json(x)
                        schema = json.dumps(adapter.json_schema())
                    if self.wrap_json:
                        def to_json(x, inner=to_json):
                            return "```json\n" + inner(x) + "\n```\n"
                        schema = "```json\n" + schema + "\n```"
                    signature = signature.with_updated_fields(
                        name,
                        desc=field.json_schema_extra.get("desc", "")
                        + (". Respond with a single JSON object. JSON Schema: " + schema),
                        format=lambda x, to_json=to_json: (x if isinstance(x, str) else to_json(x)),
                        parser=lambda x, from_json=from_json: from_json(_unwrap_json(x, from_json)),
                        schema=schema,
                        type_=type_,
                    )
            else:  # If input field
                is_json = False
                def format_(x):
                    return x if isinstance(x, str) else str(x)
                if type_ in (List[str], list[str], Tuple[str], tuple[str]):
                    format_ = passages2text
                # Special formatting for lists of known types. Maybe the output fields sohuld have this too?
                elif typing.get_origin(type_) in (List, list, Tuple, tuple):
                    (inner_type,) = typing.get_args(type_)
                    if inspect.isclass(inner_type) and issubclass(inner_type, pydantic.BaseModel):
                        def format_(x):
                            return x if isinstance(x, str) else "[" + ",".join(i.model_dump_json() for i in x) + "]"
                    else:
                        def format_(x):
                            return x if isinstance(x, str) else json.dumps(x)
                    is_json = True
                elif inspect.isclass(type_) and issubclass(type_, pydantic.BaseModel):
                    def format_(x):
                        return x if isinstance(x, str) else x.model_dump_json()
                    is_json = True
                if self.wrap_json and is_json:
                    def format_(x, inner=format_):
                        return x if isinstance(x, str) else "```json\n" + inner(x) + "\n```\n"
                signature = signature.with_updated_fields(name, format=format_)

        return signature

    def forward(self, **kwargs) -> dspy.Prediction:
        modified_kwargs = kwargs.copy()
        # We have to re-prepare the signature on every forward call, because the base
        # signature might have been modified by an optimizer or something like that.
        signature = self._prepare_signature()
        for try_i in range(self.max_retries):
            result = self.predictor(**modified_kwargs, new_signature=signature)
            errors = {}
            parsed_results = []
            # Parse the outputs
            for completion in result.completions:
                parsed = {}
                for name, field in signature.output_fields.items():
                    try:
                        value = completion[name]
                        parser = field.json_schema_extra.get("parser", lambda x: x)
                        parsed[name] = parser(value)
                    except (pydantic.ValidationError, ValueError) as e:
                        errors[name] = self._format_error(
                            e,
                            signature.fields[name],
                            value,
                            lm_explain=try_i + 1 < self.max_retries,
                        )

                        # If we can, we add an example to the error message
                        current_desc = field.json_schema_extra.get("desc", "")
                        i = current_desc.find("JSON Schema: ")
                        if i == -1:
                            continue  # Only add examples to JSON objects
                        suffix, current_desc = current_desc[i:], current_desc[:i]
                        prefix = "You MUST use this format: "
                        if (
                            try_i + 1 < self.max_retries
                            and prefix not in current_desc
                            and (example := self._make_example(field))
                        ):
                            signature = signature.with_updated_fields(
                                name,
                                desc=current_desc + "\n" + prefix + example + "\n" + suffix,
                            )
                # No reason trying to parse the general signature, or run more completions, if we already have errors
                if errors:
                    break
                # Instantiate the actual signature with the parsed values.
                # This allow pydantic to validate the fields defined in the signature.
                try:
                    _ = self.signature(**kwargs, **parsed)
                    parsed_results.append(parsed)
                except pydantic.ValidationError as e:
                    errors["general"] = self._format_error(
                        e,
                        signature.instructions,
                        "\n\n".join(
                            "> " + field.json_schema_extra["prefix"] + " " + completion[name]
                            for name, field in signature.output_fields.items()
                        ),
                        lm_explain=try_i + 1 < self.max_retries,
                    )
            if errors:
                # Add new fields for each error
                for name, error in errors.items():
                    modified_kwargs[f"error_{name}_{try_i}"] = error
                    if name == "general":
                        error_prefix = "General:"
                    else:
                        error_prefix = signature.output_fields[name].json_schema_extra["prefix"]
                    number = "" if try_i == 0 else f" ({try_i+1})"
                    signature = signature.append(
                        f"error_{name}_{try_i}",
                        dspy.InputField(
                            prefix=f"Past Error{number} in {error_prefix}",
                            desc="An error to avoid in the future",
                        ),
                    )
            else:
                # If there are no errors, we return the parsed results
                return Prediction.from_completions(
                    {key: [r[key] for r in parsed_results] for key in signature.output_fields},
                )
        raise ValueError(
            "Too many retries trying to get the correct output format. " + "Try simplifying the requirements.",
            errors,
        )


def _func_to_signature(func):
    """Make a dspy.Signature based on a function definition."""
    sig = inspect.signature(func)
    annotations = typing.get_type_hints(func, include_extras=True)
    output_key = func.__name__
    instructions = func.__doc__
    fields = {}

    # Input fields
    for param in sig.parameters.values():
        if param.name == "self":
            continue
        # We default to str as the type of the input
        annotation = annotations.get(param.name, str)
        kwargs = {}
        if typing.get_origin(annotation) is Annotated:
            desc = next((arg for arg in typing.get_args(annotation) if isinstance(arg, str)), None)
            if desc is not None:
                kwargs["desc"] = desc
        fields[param.name] = (annotation, dspy.InputField(**kwargs))

    # Output field
    kwargs = {}
    annotation = annotations.get("return", str)
    if typing.get_origin(annotation) is Annotated:
        desc = next((arg for arg in typing.get_args(annotation) if isinstance(arg, str)), None)
        if desc is not None:
            kwargs["desc"] = desc
    fields[output_key] = (annotation, dspy.OutputField(**kwargs))

    return dspy.Signature(fields, instructions)


def _unwrap_json(output, from_json: Callable[[str], Union[pydantic.BaseModel, str, None]]):
    try:
        parsing_result = from_json(output)
        if isinstance(parsing_result, pydantic.BaseModel):
            return parsing_result.model_dump_json()
        else:
            return output
    except (ValueError, pydantic.ValidationError, AttributeError):
        output = output.strip()
        if output.startswith("```"):
            if not output.startswith("```json"):
                raise ValueError("json output should start with ```json") from None
            if not output.endswith("```"):
                raise ValueError("Don't write anything after the final json ```") from None
            output = output[7:-3].strip()
        return ujson.dumps(ujson.loads(output))  # ujson is a bit more robust than the standard json


import copy

import dsp
import dspy

from .predict import Predict


class Retry(Predict):
    def __init__(self, module):
        super().__init__(module.signature)
        self.module = module
        self.original_signature = module.extended_signature if isinstance(module, dspy.ChainOfThought) else module.signature
        self.original_forward = module.forward
        self.new_signature = self._create_new_signature(self.original_signature)

    def _create_new_signature(self, signature):
        # Add "Past" input fields for each output field
        for key, value in signature.output_fields.items():
            actual_prefix = value.json_schema_extra["prefix"].split(":")[0] + ":"
            signature = signature.append(f"past_{key}", dspy.InputField(
                prefix="Previous " + actual_prefix,
                desc=f"past {actual_prefix[:-1]} with errors",
                format=value.json_schema_extra.get("format"),
            ))

        signature = signature.append("feedback", dspy.InputField(
            prefix="Instructions:",
            desc="Some instructions you must satisfy",
            format=str,
        ))

        return signature

    def forward(self, *, past_outputs, **kwargs):
        # Take into account the possible new signature, as in TypedPredictor
        new_signature = kwargs.pop("new_signature", None)
        if new_signature:
            self.original_signature = new_signature
            self.new_signature = self._create_new_signature(self.original_signature)

        # Convert the dict past_outputs={"answer": ...} to kwargs
        # {past_answer=..., ...}
        for key, value in past_outputs.items():
            past_key = f"past_{key}"
            if past_key in self.new_signature.input_fields:
                kwargs[past_key] = value
        # Tell the wrapped module to use the new signature.
        # Note: This only works if the wrapped module is a Predict or ChainOfThought.
        kwargs["new_signature"] = self.new_signature
        return self.original_forward(**kwargs)

    def __call__(self, **kwargs):
        copy.deepcopy(kwargs)
        kwargs["_trace"] = False
        kwargs.setdefault("demos", self.demos if self.demos is not None else [])

        # perform backtracking
        if dspy.settings.backtrack_to == self:
            for key, value in dspy.settings.backtrack_to_args.items():
                kwargs.setdefault(key, value)
            pred = self.forward(**kwargs)
        else:
            pred = self.module(**kwargs)

        # now pop multiple reserved keys
        # NOTE(shangyin) past_outputs seems not useful to include in demos,
        # therefore dropped
        for key in ["_trace", "demos", "signature", "new_signature", "config", "lm", "past_outputs"]:
            kwargs.pop(key, None)

        if dsp.settings.trace is not None:
            trace = dsp.settings.trace
            trace.append((self, {**kwargs}, pred))
        return pred



import os
import warnings
from typing import Any, Optional

import backoff

from dsp.modules.lm import LM
from dsp.utils.settings import settings

try:
    import premai

    premai_api_error = premai.errors.UnexpectedStatus
except ImportError:
    premai_api_error = Exception
except AttributeError:
    premai_api_error = Exception


def backoff_hdlr(details) -> None:
    """Handler for the backoff package.

    See more at: https://pypi.org/project/backoff/
    """
    print(  # noqa: T201
        "Backing off {wait:0.1f} seconds after {tries} tries calling function {target} with kwargs {kwargs}".format(
            **details,
        ),
    )


def giveup_hdlr(details) -> bool:
    """Wrapper function that decides when to give up on retry."""
    if "rate limits" in details.message:
        return False
    return True


def get_premai_api_key(api_key: Optional[str] = None) -> str:
    """Retrieve the PreMAI API key from a passed argument or environment variable."""
    api_key = api_key or os.environ.get("PREMAI_API_KEY")
    if api_key is None:
        raise RuntimeError(
            "No API key found. See the quick start guide at https://docs.premai.io/introduction to get your API key.",
        )
    return api_key


class PremAI(LM):
    def __init__(
        self,
        project_id: int,
        model: Optional[str] = None,
        api_key: Optional[str] = None,
        **kwargs: dict,
    ) -> None:
        """Parameters

        project_id: int
            "The project ID in which the experiments or deployments are carried out. can find all your projects here: https://app.premai.io/projects/"
        model: Optional[str]
            The name of model deployed on launchpad. When None, it will show 'default'
        api_key: Optional[str]
            Prem AI API key, to connect with the API. If not provided then it will check from env var by the name
                PREMAI_API_KEY
        kwargs: Optional[dict] For any additional parameters
        """
        self.model = "default" if model is None else model
        super().__init__(self.model)
        if premai_api_error == Exception:
            raise ImportError(
                "Not loading Prem AI because it is not installed. Install it with `pip install premai`.",
            )

        self.project_id = project_id

        api_key = get_premai_api_key(api_key=api_key)
        self.client = premai.Prem(api_key=api_key)
        self.provider = "premai"
        self.history: list[dict[str, Any]] = []
        self.kwargs = kwargs if kwargs else {}

    @property
    def _default_params(self) -> dict[str, Any]:
        default_kwargs = {
            "temperature": None,
            "max_tokens": None,
            "system_prompt": None,
            "repositories": None,
        }

        if self.model != "default":
            default_kwargs["model_name"] = self.model

        return default_kwargs

    def _get_all_kwargs(self, **kwargs) -> dict[str, Any]:
        kwargs_to_ignore = [
            "top_p",
            "tools",
            "frequency_penalty",
            "presence_penalty",
            "logit_bias",
            "stop",
            "seed",
        ]
        keys_to_remove = []
        kwargs = {**kwargs, **self.kwargs}

        for key in kwargs:
            if key in kwargs_to_ignore:
                warnings.warn(f"WARNING: Parameter {key} is not supported in kwargs.", stacklevel=2)
                keys_to_remove.append(key)

        for key in keys_to_remove:
            kwargs.pop(key)

        all_kwargs = {**self._default_params, **kwargs}
        for key in list(self._default_params.keys()):
            if all_kwargs.get(key) is None or all_kwargs.get(key) == "":
                all_kwargs.pop(key, None)
        return all_kwargs

    def basic_request(self, prompt, **kwargs) -> list[str]:
        """Handles retrieval of completions from Prem AI whilst handling API errors."""
        all_kwargs = self._get_all_kwargs(**kwargs)
        if "template_id" not in all_kwargs:
            messages = [{"role": "user", "content": prompt}]
        else:
            template_id = all_kwargs["template_id"]
            if (template_id is None) or (template_id == ""):
                raise ValueError("Templates can not be None or ''")
            if "params" not in all_kwargs:
                raise KeyError(
                    "Keyword argument: params must be present if template_id is present",
                )

            params = all_kwargs["params"]
            if not isinstance(params, dict):
                raise TypeError("params must be a dictionary")

            messages = [
                {
                    "role": "user",
                    "template_id": template_id,
                    "params": params,
                },
            ]
            kwargs["template_id"] = all_kwargs.get("template_id", None)
            kwargs["params"] = all_kwargs.get("params", None)

            all_kwargs.pop("template_id")
            all_kwargs.pop("params")

        kwargs = {**kwargs, **all_kwargs}
        response = self.client.chat.completions.create(
            project_id=self.project_id,
            messages=messages,
            **all_kwargs,
        )
        content = response.choices[0].message.content
        if not content:
            raise premai_api_error("ChatResponse is none")

        output_text = content or ""
        document_chunks = (
            None if response.document_chunks is None else [chunk.to_dict() for chunk in response.document_chunks]
        )

        self.history.append(
            {
                "prompt": prompt,
                "response": content,
                "document_chunks": document_chunks,
                "kwargs": kwargs,
            },
        )
        return [output_text]

    @backoff.on_exception(
        backoff.expo,
        (premai_api_error),
        max_time=settings.backoff_time,
        on_backoff=backoff_hdlr,
        giveup=giveup_hdlr,
    )
    def request(self, prompt, **kwargs) -> str:
        """Handles retrieval of completions from Prem AI whilst handling API errors."""
        return self.basic_request(prompt=prompt, **kwargs)

    def __call__(self, prompt, **kwargs):
        return self.request(prompt, **kwargs)



import functools
import json
from typing import Literal, Optional

import openai

from dsp.modules.cache_utils import CacheMemory, NotebookCacheMemory, cache_turn_on
from dsp.modules.gpt3 import GPT3

try:
    import openai.error
    from openai.openai_object import OpenAIObject
    ERRORS = (openai.error.RateLimitError, openai.error.ServiceUnavailableError, openai.error.APIError)
except Exception:
    ERRORS = (openai.RateLimitError, openai.APIError)
    OpenAIObject = dict


def backoff_hdlr(details):
    """Handler from https://pypi.org/project/backoff/"""
    print(
        "Backing off {wait:0.1f} seconds after {tries} tries "
        "calling function {target} with kwargs "
        "{kwargs}".format(**details),
    )

class Databricks(GPT3):
    """Wrapper around DSPy's OpenAI Wrapper. Supports Databricks Model Serving Endpoints for OpenAI SDK on both Chat, Completions, and Embeddings models.

    Args:
        model (str, required): Databricks-hosted LLM model to use.
        api_key (Optional[str], optional): Databricks authentication token. Defaults to None.
        api_base (Optional[str], optional): Databricks model serving endpoint. Defaults to None.
        model_type (Literal["chat", "text"], optional): The type of model that was specified. Mainly to decide the optimal prompting strategy. Defaults to "text".
        **kwargs: Additional arguments to pass to the OpenAI API provider.
    """

    def __init__(
        self,
        model: str,
        api_key: Optional[str] = None,
        api_base: Optional[str] = None,
        model_type: Literal["chat", "text", "embeddings"] = None,
        **kwargs,
    ):
        super().__init__(
            model=model,
            api_key=api_key,
            api_provider="openai",
            api_base=api_base,
            model_type=model_type,
            **kwargs,
        )

        self.kwargs.pop('frequency_penalty', None)
        self.kwargs.pop('presence_penalty', None)    

    def basic_request(self, prompt: str, **kwargs):
        raw_kwargs = kwargs

        kwargs = {**self.kwargs, **kwargs}
        if self.model_type == "chat":
            kwargs["messages"] = [{"role": "system", "content": "You are a helpful assistant."}, {"role": "user", "content": prompt}]
            kwargs = {"stringify_request": json.dumps(kwargs)}
            response = custom_client_chat_request(**kwargs).json()
            response = json.loads(response)
        else:
            kwargs["prompt"] = prompt
            response = custom_client_completions_request(**kwargs).json()    
            response = json.loads(response)
        
        history = {
            "prompt": prompt,
            "response": response,
            "kwargs": kwargs,
            "raw_kwargs": raw_kwargs,
        }
        self.history.append(history)
        return response
    
    def embeddings(self, prompt: str, **kwargs):
        kwargs = {**self.kwargs, **kwargs}
        kwargs["input"] = prompt
        kwargs.pop('temperature', None)
        kwargs.pop('max_tokens', None)
        kwargs.pop('top_p', None)
        kwargs.pop('n', None)
        response = custom_client_embeddings_request(**kwargs).json()    
        response = json.loads(response)
        embeddings = [cur_obj['embedding'] for cur_obj in response['data']][0]
        return embeddings
    
    def __call__(self, prompt: str, **kwargs):
        if self.model_type == "embeddings":
            return self.embeddings(prompt, **kwargs)
        else:
            return super().__call__(prompt, **kwargs)

def create_custom_client():
    from openai import OpenAI
    
    client = OpenAI(api_key=openai.api_key, base_url=openai.base_url)
    return client

def custom_client_chat_request(**kwargs):
    return cached_custom_client_chat_request_v2_wrapped(**kwargs)

def custom_client_embeddings_request(**kwargs):
    return cached_custom_client_embeddings_request_v2_wrapped(**kwargs)

def custom_client_completions_request(**kwargs):
    return cached_custom_client_completions_request_v2_wrapped(**kwargs)


@CacheMemory.cache
def cached_custom_client_chat_request_v2(**kwargs):
    client = create_custom_client()
    return client.chat.completions.create(**kwargs)

@functools.lru_cache(maxsize=None if cache_turn_on else 0)
@NotebookCacheMemory.cache
def cached_custom_client_chat_request_v2_wrapped(**kwargs):
    if "stringify_request" in kwargs:
        kwargs = json.loads(kwargs["stringify_request"])
    return cached_custom_client_chat_request_v2(**kwargs)

@CacheMemory.cache
def cached_custom_client_completions_request_v2(**kwargs):
    client = create_custom_client()
    return client.completions.create(**kwargs)

@functools.lru_cache(maxsize=None if cache_turn_on else 0)
@NotebookCacheMemory.cache
def cached_custom_client_completions_request_v2_wrapped(**kwargs):
    return cached_custom_client_completions_request_v2(**kwargs)

@CacheMemory.cache
def cached_custom_client_embeddings_request_v2(**kwargs):
    client = create_custom_client()
    return client.embeddings.create(**kwargs)

@functools.lru_cache(maxsize=None if cache_turn_on else 0)
@NotebookCacheMemory.cache
def cached_custom_client_embeddings_request_v2_wrapped(**kwargs):
    return cached_custom_client_embeddings_request_v2(**kwargs)



import math
import random
import sys
import textwrap
from collections import defaultdict
from typing import Any

import optuna

import dsp
import dspy
from dspy.evaluate.evaluate import Evaluate
from dspy.signatures import Signature
from dspy.signatures.signature import signature_to_template
from dspy.teleprompt import BootstrapFewShot
from dspy.teleprompt.teleprompt import Teleprompter

"""
USAGE SUGGESTIONS:

The following code can be used to compile a optimized signature teleprompter using MIPRO, and evaluate it on an end task:

``` python
from dspy.teleprompt import MIPRO

teleprompter = MIPRO(prompt_model=prompt_model, task_model=task_model, metric=metric, num_candidates=10, init_temperature=1.0)
kwargs = dict(num_threads=NUM_THREADS, display_progress=True, display_table=0)
compiled_prompt_opt = teleprompter.compile(program, trainset=trainset[:TRAIN_NUM], num_trials=100, max_bootstrapped_demos=3, max_labeled_demos=5, eval_kwargs=kwargs)
eval_score = evaluate(compiled_prompt_opt, devset=evalset[:EVAL_NUM], **kwargs)
```

Note that this teleprompter takes in the following parameters:

* prompt_model: The model used for prompt generation. When unspecified, defaults to the model set in settings (i.e., dspy.settings.configure(lm=task_model)).
* task_model: The model used for running your task. When unspecified, defaults to the model set in settings (i.e., dspy.settings.configure(lm=task_model)).
* metric: The task metric used for optimization.
* num_candidates: The number of new prompts and sets of fewshot examples to generate and evaluate. Default=10.
* init_temperature: The temperature used to generate new prompts. Higher roughly equals more creative. Default=1.0.
* verbose: Tells the method whether or not to print intermediate steps.
* track_stats: Tells the method whether or not to track statistics about the optimization process.
                If True, the method will track a dictionary with a key corresponding to the trial number, 
                and a value containing a dict with the following keys:
                    * program: the program being evaluated at a given trial
                    * score: the last average evaluated score for the program
                    * pruned: whether or not this program was pruned
                This information will be returned as attributes of the best program.
"""


class BasicGenerateInstruction(Signature):
    """You are an instruction optimizer for large language models. I will give you a ``signature`` of fields (inputs and outputs) in English. Your task is to propose an instruction that will lead a good language model to perform the task well. Don't be afraid to be creative."""

    basic_instruction = dspy.InputField(desc="The initial instructions before optimization")
    proposed_instruction = dspy.OutputField(desc="The improved instructions for the language model")
    proposed_prefix_for_output_field = dspy.OutputField(
        desc="The string at the end of the prompt, which will help the model start solving the task",
    )


class BasicGenerateInstructionWithDataObservations(Signature):
    """You are an instruction optimizer for large language models. I will give you a ``signature`` of fields (inputs and outputs) in English.  I will also give you some ``observations`` I have made about the dataset and task. Your task is to propose an instruction that will lead a good language model to perform the task well. Don't be afraid to be creative."""

    basic_instruction = dspy.InputField(desc="The initial instructions before optimization")
    observations = dspy.InputField(desc="Observations about the dataset and task")
    proposed_instruction = dspy.OutputField(desc="The improved instructions for the language model")
    proposed_prefix_for_output_field = dspy.OutputField(
        desc="The string at the end of the prompt, which will help the model start solving the task",
    )


class BasicGenerateInstructionWithExamples(dspy.Signature):
    """You are an instruction optimizer for large language models. I will give you a ``signature`` of fields (inputs and outputs) in English. Specifically, I will also provide you with the current ``basic instruction`` that is being used for this task. I will also provide you with some ``examples`` of the expected inputs and outputs.

    Your task is to propose an instruction that will lead a good language model to perform the task well. Don't be afraid to be creative."""

    # attempted_instructions = dspy.InputField(format=str, desc="Previously attempted task instructions, along with their resulting validation score, and an example of the instruction in use on a sample from our dataset.")
    basic_instruction = dspy.InputField(desc="The initial instructions before optimization")
    # examples = dspy.InputField(format=dsp.passages2text, desc="Example(s) of the task")
    examples = dspy.InputField(format=dsp.passages2text, desc="Example(s) of the task")
    proposed_instruction = dspy.OutputField(desc="The improved instructions for the language model")
    proposed_prefix_for_output_field = dspy.OutputField(
        desc="The string at the end of the prompt, which will help the model start solving the task",
    )


class BasicGenerateInstructionWithExamplesAndDataObservations(dspy.Signature):
    """You are an instruction optimizer for large language models. I will give you a ``signature`` of fields (inputs and outputs) in English. Specifically, I will give you some ``observations`` I have made about the dataset and task, along with some ``examples`` of the expected inputs and outputs. I will also provide you with the current ``basic instruction`` that is being used for this task.

    Your task is to propose a new improved instruction and prefix for the output field that will lead a good language model to perform the task well. Don't be afraid to be creative."""

    observations = dspy.InputField(desc="Observations about the dataset and task")
    examples = dspy.InputField(format=dsp.passages2text, desc="Example(s) of the task")
    basic_instruction = dspy.InputField(desc="The initial instructions before optimization")
    proposed_instruction = dspy.OutputField(desc="The improved instructions for the language model")
    proposed_prefix_for_output_field = dspy.OutputField(
        desc="The string at the end of the prompt, which will help the model start solving the task",
    )


class ObservationSummarizer(dspy.Signature):
    """Given a series of observations I have made about my dataset, please summarize them into a brief 2-3 sentence summary which highlights only the most important details."""

    observations = dspy.InputField(desc="Observations I have made about my dataset")
    summary = dspy.OutputField(
        desc="Two to Three sentence summary of only the most significant highlights of my observations",
    )


class DatasetDescriptor(dspy.Signature):
    (
        """Given several examples from a dataset please write observations about trends that hold for most or all of the samples. """
        """Some areas you may consider in your observations: topics, content, syntax, conciceness, etc. """
        """It will be useful to make an educated guess as to the nature of the task this dataset will enable. Don't be afraid to be creative"""
    )

    examples = dspy.InputField(desc="Sample data points from the dataset")
    observations = dspy.OutputField(desc="Somethings that holds true for most or all of the data you observed")


class DatasetDescriptorWithPriorObservations(dspy.Signature):
    (
        """Given several examples from a dataset please write observations about trends that hold for most or all of the samples. """
        """I will also provide you with a few observations I have already made.  Please add your own observations or if you feel the observations are comprehensive say 'COMPLETE' """
        """Some areas you may consider in your observations: topics, content, syntax, conciceness, etc. """
        """It will be useful to make an educated guess as to the nature of the task this dataset will enable. Don't be afraid to be creative"""
    )

    examples = dspy.InputField(desc="Sample data points from the dataset")
    prior_observations = dspy.InputField(desc="Some prior observations I made about the data")
    observations = dspy.OutputField(
        desc="Somethings that holds true for most or all of the data you observed or COMPLETE if you have nothing to add",
    )


class MIPRO(Teleprompter):
    def __init__(
        self,
        metric,
        prompt_model=None,
        task_model=None,
        teacher_settings={},
        num_candidates=10,
        init_temperature=1.0,
        verbose=False,
        track_stats=True,
        view_data_batch_size=10,
    ):
        self.num_candidates = num_candidates
        self.metric = metric
        self.init_temperature = init_temperature
        self.prompt_model = prompt_model if prompt_model is not None else dspy.settings.lm
        self.task_model = task_model if task_model is not None else dspy.settings.lm
        self.verbose = verbose
        self.track_stats = track_stats
        self.teacher_settings = teacher_settings
        self.view_data_batch_size = view_data_batch_size

    def _print_full_program(self, program):
        for i, predictor in enumerate(program.predictors()):
            if self.verbose:
                print(f"Predictor {i}")
            if self.verbose:
                print(f"i: {self._get_signature(predictor).instructions}")
            *_, last_field = self._get_signature(predictor).fields.values()
            if self.verbose:
                print(f"p: {last_field.json_schema_extra['prefix']}")
            if self.verbose:
                print("\n")

    def _print_model_history(self, model, n=1):
        if self.verbose:
            print(f"Model ({model}) History:")
        model.inspect_history(n=n)

    def _observe_data(self, trainset, max_iterations=10):
        upper_lim = min(len(trainset), self.view_data_batch_size)
        observation = dspy.Predict(DatasetDescriptor, n=1, temperature=1.0)(examples=(trainset[0:upper_lim].__repr__()))
        observations = observation["observations"]

        skips = 0
        iterations = 0
        for b in range(self.view_data_batch_size, len(trainset), self.view_data_batch_size):
            upper_lim = min(len(trainset), b + self.view_data_batch_size)
            output = dspy.Predict(DatasetDescriptorWithPriorObservations, n=1, temperature=1.0)(
                prior_observations=observations,
                examples=(trainset[b:upper_lim].__repr__()),
            )
            iterations += 1
            if len(output["observations"]) >= 8 and output["observations"][:8].upper() == "COMPLETE":
                skips += 1
                if skips >= 5:
                    break
                continue
            if iterations >= max_iterations:
                break
            observations += output["observations"]

        summary = dspy.Predict(ObservationSummarizer, n=1, temperature=1.0)(observations=observations)

        return summary.summary

    def _create_example_string(self, fields, example):
        # Building the output string
        output = []
        for field in fields:
            name = field.name
            separator = field.separator
            input_variable = field.input_variable

            # Determine the value from input_data or prediction_data
            value = example.get(input_variable)

            # Construct the string for the current field
            field_str = f"{name}{separator}{value}"
            output.append(field_str)

        # Joining all the field strings
        return "\n".join(output)

    def _get_signature(self, predictor):
        if hasattr(predictor, "extended_signature"):
            return predictor.extended_signature
        elif hasattr(predictor, "signature"):
            return predictor.signature
        return None

    def _set_signature(self, predictor, updated_signature):
        if hasattr(predictor, "extended_signature"):
            predictor.extended_signature = updated_signature
        elif hasattr(predictor, "signature"):
            predictor.signature = updated_signature

    def _generate_first_N_candidates(  # noqa: N802
        self,
        module: dspy.Module,
        N: int,  # noqa: N803
        view_data: bool,
        view_examples: bool,
        demo_candidates: dict,
        devset,
    ) -> tuple[dict, dict]:
        candidates = {}
        evaluated_candidates = defaultdict(dict)

        if view_data:
            # Create data observations
            self.observations = None
            with dspy.settings.context(lm=self.prompt_model):
                self.observations = self._observe_data(devset).replace("Observations:", "").replace("Summary:", "")

        if view_examples:
            example_sets = {}
            for predictor in module.predictors():
                # Get all augmented examples
                example_set = {}
                all_sets_of_examples = demo_candidates[id(predictor)]  # Get all generated sets of examples
                for example_set_i, set_of_examples in enumerate(all_sets_of_examples):
                    if example_set_i != 0:  # Skip the no examples case
                        for example in set_of_examples:  # Get each individual example in the set
                            if "augmented" in example and example["augmented"]:
                                if example_set_i not in example_set:
                                    example_set[example_set_i] = []
                                fields_to_use = signature_to_template(predictor.signature).fields
                                _input_variable_names = list(self._get_signature(predictor).input_fields.keys())
                                example_string = self._create_example_string(fields_to_use, example)
                                example_set[example_set_i].append(example_string)
                        example_sets[id(predictor)] = example_set
                    else:
                        example_set[example_set_i] = []
                        example_sets[id(predictor)] = example_set

        # Seed the prompt optimizer zero shot with just the instruction, generate BREADTH new prompts
        for predictor in module.predictors():
            basic_instruction = None
            basic_prefix = None
            basic_instruction = self._get_signature(predictor).instructions
            *_, last_field = self._get_signature(predictor).fields.values()
            basic_prefix = last_field.json_schema_extra["prefix"]
            with dspy.settings.context(lm=self.prompt_model):
                # Data & Examples
                if view_data and view_examples:
                    if 1 not in example_sets[id(predictor)]:
                        raise ValueError("No examples found for the given predictor")
                    instruct = None
                    for i in range(1, self.num_candidates):
                        new_instruct = dspy.Predict(
                            BasicGenerateInstructionWithExamplesAndDataObservations,
                            n=1,
                            temperature=self.init_temperature,
                        )(
                            basic_instruction=basic_instruction,
                            observations=self.observations,
                            examples=example_sets[id(predictor)][i],
                        )
                        if not instruct:
                            instruct = new_instruct
                        else:
                            instruct.completions.proposed_instruction.extend(
                                new_instruct.completions.proposed_instruction,
                            )
                            instruct.completions.proposed_prefix_for_output_field.extend(
                                new_instruct.completions.proposed_prefix_for_output_field,
                            )
                # Just data
                elif view_data:
                    instruct = dspy.Predict(
                        BasicGenerateInstructionWithDataObservations,
                        n=N - 1,
                        temperature=self.init_temperature,
                    )(basic_instruction=basic_instruction, observations=self.observations)
                # Just examples
                elif view_examples:
                    instruct = None
                    for i in range(1, self.num_candidates):  # Note: skip over the first example set which is empty
                        new_instruct = dspy.Predict(
                            BasicGenerateInstructionWithExamples,
                            n=1,
                            temperature=self.init_temperature,
                        )(
                            basic_instruction=basic_instruction,
                            examples=example_sets[id(predictor)][i],
                        )
                        if not instruct:
                            instruct = new_instruct
                        else:
                            instruct.completions.proposed_instruction.extend(
                                new_instruct.completions.proposed_instruction,
                            )
                            instruct.completions.proposed_prefix_for_output_field.extend(
                                new_instruct.completions.proposed_prefix_for_output_field,
                            )
                # Neither
                else:
                    instruct = dspy.Predict(BasicGenerateInstruction, n=N - 1, temperature=self.init_temperature)(
                        basic_instruction=basic_instruction,
                    )

            # Add in our initial prompt as a candidate as well
            instruct.completions.proposed_instruction.insert(0, basic_instruction)
            instruct.completions.proposed_prefix_for_output_field.insert(0, basic_prefix)
            candidates[id(predictor)] = instruct.completions
            evaluated_candidates[id(predictor)] = {}

        if self.verbose:
            self._print_model_history(self.prompt_model)

        return candidates, evaluated_candidates

    def compile(
        self,
        student: dspy.Program,
        *,
        trainset: list[dspy.Example],
        num_trials: int,
        max_bootstrapped_demos: int,
        max_labeled_demos: int,
        eval_kwargs: dict[str, Any],
        seed=42,
        view_data=True,
        view_examples=True,
        requires_permission_to_run=True,
    ) -> dspy.Program:
        # Define ANSI escape codes for colors
        RED = "\033[91m"
        YELLOW = "\033[93m"
        BLUE = "\033[94m"
        BOLD = "\033[1m"
        ENDC = "\033[0m"  # Resets the color to default

        random.seed(seed)

        estimated_task_model_calls_wo_module_calls = len(trainset) * num_trials  # M * T * P
        estimated_prompt_model_calls = 10 + self.num_candidates * len(
            student.predictors(),
        )  # num data summary calls + N * P

        user_message = textwrap.dedent(f"""\
            {YELLOW}{BOLD}WARNING: Projected Language Model (LM) Calls{ENDC}

            Please be advised that based on the parameters you have set, the maximum number of LM calls is projected as follows:

            {YELLOW}- Task Model: {BLUE}{BOLD}{len(trainset)}{ENDC}{YELLOW} examples in dev set * {BLUE}{BOLD}{num_trials}{ENDC}{YELLOW} trials * {BLUE}{BOLD}# of LM calls in your program{ENDC}{YELLOW} = ({BLUE}{BOLD}{estimated_task_model_calls_wo_module_calls} * # of LM calls in your program{ENDC}{YELLOW}) task model calls{ENDC}
            {YELLOW}- Prompt Model: # data summarizer calls (max {BLUE}{BOLD}10{ENDC}{YELLOW}) + {BLUE}{BOLD}{self.num_candidates}{ENDC}{YELLOW} * {BLUE}{BOLD}{len(student.predictors())}{ENDC}{YELLOW} lm calls in program = {BLUE}{BOLD}{estimated_prompt_model_calls}{ENDC}{YELLOW} prompt model calls{ENDC}

            {YELLOW}{BOLD}Estimated Cost Calculation:{ENDC}

            {YELLOW}Total Cost = (Number of calls to task model * (Avg Input Token Length per Call * Task Model Price per Input Token + Avg Output Token Length per Call * Task Model Price per Output Token) 
                        + (Number of calls to prompt model * (Avg Input Token Length per Call * Task Prompt Price per Input Token + Avg Output Token Length per Call * Prompt Model Price per Output Token).{ENDC}

            For a preliminary estimate of potential costs, we recommend you perform your own calculations based on the task
            and prompt models you intend to use. If the projected costs exceed your budget or expectations, you may consider:

            {YELLOW}- Reducing the number of trials (`num_trials`), the size of the trainset, or the number of LM calls in your program.{ENDC}
            {YELLOW}- Using a cheaper task model to optimize the prompt.{ENDC}""")

        user_confirmation_message = textwrap.dedent(f"""\
            To proceed with the execution of this program, please confirm by typing {BLUE}'y'{ENDC} for yes or {BLUE}'n'{ENDC} for no.

            If you would like to bypass this confirmation step in future executions, set the {YELLOW}`requires_permission_to_run`{ENDC} flag to {YELLOW}`False` when calling compile.{ENDC}

            {YELLOW}Awaiting your input...{ENDC}
        """)

        print(f"""{RED}{BOLD}WARNING: MIPRO has been deprecated and replaced with MIPROv2.  MIPRO will be removed in a future release. {ENDC}""")
        print(user_message)

        sys.stdout.flush()  # Flush the output buffer to force the message to print

        run = True
        if requires_permission_to_run:
            print(user_confirmation_message)
            user_input = input("Do you wish to continue? (y/n): ").strip().lower()
            if user_input != "y":
                print("Compilation aborted by the user.")
                run = False

        if run:
            # Set up program and evaluation function
            module = student.deepcopy()
            evaluate = Evaluate(devset=trainset, metric=self.metric, **eval_kwargs)

            # In the case where the bootstrapped and labeled demos are set to 0, we'll still bootstrap examples to use in our meta prompt
            if (
                max_bootstrapped_demos == 0 and max_labeled_demos == 0
            ):  # TODO: address case when max_bootstrapped alone is 0
                max_bootstrapped_demos_for_candidate_gen = 1
                max_labeled_demos_for_candidate_gen = 1  # TODO: this might only need to be 0
            else:
                max_bootstrapped_demos_for_candidate_gen = max_bootstrapped_demos
                max_labeled_demos_for_candidate_gen = max_labeled_demos

            # Generate N few shot example sets
            demo_candidates = {}
            for i in range(self.num_candidates):
                if i == 0:  # Story empty set of demos as default for index 0
                    for module_p in module.predictors():
                        if id(module_p) not in demo_candidates:
                            demo_candidates[id(module_p)] = []
                        demo_candidates[id(module_p)].append([])
                else:
                    if self.verbose:
                        print(f"Creating basic bootstrap: {i}/{self.num_candidates-1}")

                    # Create a new basic bootstrap few - shot program .
                    rng = random.Random(i)
                    shuffled_trainset = trainset[:]  # Create a copy of devset
                    rng.shuffle(shuffled_trainset)  # Shuffle the copy
                    tp = BootstrapFewShot(
                        metric=self.metric,
                        max_bootstrapped_demos=max_bootstrapped_demos_for_candidate_gen,
                        max_labeled_demos=max_labeled_demos_for_candidate_gen,
                        teacher_settings=self.teacher_settings,
                    )
                    candidate_program = tp.compile(student=module.deepcopy(), trainset=shuffled_trainset)

                    # Store the candidate demos
                    for module_p, candidate_p in zip(module.predictors(), candidate_program.predictors()):
                        if id(module_p) not in demo_candidates:
                            demo_candidates[id(module_p)] = []
                        demo_candidates[id(module_p)].append(candidate_p.demos)

            # Generate N candidate prompts
            instruction_candidates, _ = self._generate_first_N_candidates(
                module,
                self.num_candidates,
                view_data,
                view_examples,
                demo_candidates,
                trainset,
            )

            # Reset demo_candidates to None for our optimization if the user asked for no fewshot examples
            if max_bootstrapped_demos == 0 and max_labeled_demos == 0:
                demo_candidates = None

            # Initialize variables to store the best program and its score
            best_score = float("-inf")
            best_program = None
            trial_num = 0

            trial_logs = {}

            # Define our trial objective
            def create_objective(baseline_program, instruction_candidates, demo_candidates, evaluate, trainset):
                def objective(trial):
                    nonlocal best_program, best_score, trial_num, trial_logs  # Allow access to the outer variables
                    candidate_program = baseline_program.deepcopy()

                    # Suggest the instruction to use for our predictor
                    print(f"Starting trial #{trial_num}")
                    trial_logs[trial_num] = {}

                    for p_old, p_new in zip(baseline_program.predictors(), candidate_program.predictors()):
                        # Get instruction candidates for our given predictor
                        p_instruction_candidates = instruction_candidates[id(p_old)]
                        if demo_candidates:
                            p_demo_candidates = demo_candidates[id(p_old)]

                        # Suggest the index of the instruction candidate to use in our trial
                        instruction_idx = trial.suggest_categorical(
                            f"{id(p_old)}_predictor_instruction",
                            range(len(p_instruction_candidates)),
                        )
                        if demo_candidates:
                            demos_idx = trial.suggest_categorical(
                                f"{id(p_old)}_predictor_demos",
                                range(len(p_demo_candidates)),
                            )
                        trial_logs[trial_num][f"{id(p_old)}_predictor_instruction"] = instruction_idx
                        if demo_candidates:
                            trial_logs[trial_num][f"{id(p_old)}_predictor_demos"] = demos_idx

                        # Get the selected instruction candidate
                        selected_candidate = p_instruction_candidates[instruction_idx]
                        selected_instruction = selected_candidate.proposed_instruction.strip('"').strip()
                        selected_prefix = selected_candidate.proposed_prefix_for_output_field.strip('"').strip()

                        # Use this candidates in our program
                        *_, last_field = self._get_signature(p_new).fields.keys()
                        updated_signature = (
                            self._get_signature(p_new)
                            .with_instructions(selected_instruction)
                            .with_updated_fields(last_field, prefix=selected_prefix)
                        )
                        self._set_signature(p_new, updated_signature)

                        # Get the selected demos
                        if demo_candidates:
                            selected_demos = p_demo_candidates[demos_idx]

                        # Use these demos in our program
                        if demo_candidates:
                            p_new.demos = selected_demos

                    if self.verbose:
                        print("Evaling the following program:")
                    if self.verbose:
                        self._print_full_program(candidate_program)
                    trial_logs[trial_num]["program"] = candidate_program

                    # Evaluate with the new prompts
                    total_score = 0
                    batch_size = 100
                    num_batches = math.ceil(len(trainset) / batch_size)

                    for i in range(num_batches):
                        start_index = i * batch_size
                        end_index = min((i + 1) * batch_size, len(trainset))
                        split_trainset = trainset[start_index:end_index]
                        split_score = evaluate(candidate_program, devset=split_trainset, display_table=0)
                        if self.verbose:
                            print(f"{i}st split score: {split_score}")

                        total_score += split_score * len(split_trainset)
                        curr_weighted_avg_score = total_score / min((i + 1) * 100, len(trainset))
                        if self.verbose:
                            print(f"curr average score: {curr_weighted_avg_score}")

                        trial.report(curr_weighted_avg_score, i)

                        # Handle pruning based on the intermediate value.
                        if trial.should_prune():
                            print("Trial pruned.")
                            trial_logs[trial_num]["score"] = curr_weighted_avg_score
                            trial_logs[trial_num]["pruned"] = True
                            trial_num += 1
                            raise optuna.TrialPruned()

                    if self.verbose:
                        print(f"Fully evaled score: {curr_weighted_avg_score}")
                    if self.verbose:
                        self._print_model_history(self.task_model, n=1)
                    score = curr_weighted_avg_score

                    trial_logs[trial_num]["score"] = curr_weighted_avg_score
                    trial_logs[trial_num]["pruned"] = False

                    # Update the best program if the current score is better
                    if score > best_score:
                        best_score = score
                        best_program = candidate_program.deepcopy()

                    trial_num += 1

                    return score

                return objective

            # Run the trial
            objective_function = create_objective(module, instruction_candidates, demo_candidates, evaluate, trainset)
            sampler = optuna.samplers.TPESampler(seed=seed)
            study = optuna.create_study(direction="maximize", sampler=sampler)
            _score = study.optimize(objective_function, n_trials=num_trials)

            if best_program is not None and self.track_stats:
                best_program.trial_logs = trial_logs

            print(f"Returning {best_program} from continue_program")
            return best_program
        return None



from anyio import CapacityLimiter
import asyncer


_limiter = None


def get_async_max_workers():
    import dspy

    return dspy.settings.async_max_workers


def get_limiter():
    async_max_workers = get_async_max_workers()

    global _limiter
    if _limiter is None:
        _limiter = CapacityLimiter(async_max_workers)
    elif _limiter.total_tokens != async_max_workers:
        _limiter.total_tokens = async_max_workers

    return _limiter


def asyncify(program):
    import dspy
    import threading

    assert threading.get_ident() == dspy.settings.main_tid, "asyncify can only be called from the main thread"

    def wrapped(*args, **kwargs):
        thread_stacks = dspy.settings.stack_by_thread
        current_thread_id = threading.get_ident()
        creating_new_thread = current_thread_id not in thread_stacks

        assert creating_new_thread
        thread_stacks[current_thread_id] = list(dspy.settings.main_stack)

        try:
            return program(*args, **kwargs)
        finally:
            del thread_stacks[threading.get_ident()]

    return asyncer.asyncify(wrapped, abandon_on_cancel=True, limiter=get_limiter())



import functools
import json
import logging
from typing import Any, Literal, Optional, cast
import backoff
import httpx
try:
    """
    If there is any error in the langfuse configuration, it will turn to request the real address(openai or azure endpoint)
    """
    import langfuse
    from langfuse.openai import openai
    logging.info(f"You are using Langfuse,version{langfuse.__version__}")
except:
    import openai

from dsp.modules.cache_utils import CacheMemory, NotebookCacheMemory, cache_turn_on
from dsp.modules.lm import LM
from dsp.utils.settings import settings

try:
    OPENAI_LEGACY = int(openai.version.__version__[0]) == 0
except Exception:
    OPENAI_LEGACY = True

try:
    import openai.error
    from openai.openai_object import OpenAIObject

    ERRORS = (openai.error.RateLimitError,)
except Exception:
    ERRORS = (openai.RateLimitError,)
    OpenAIObject = dict


def backoff_hdlr(details):
    """Handler from https://pypi.org/project/backoff/"""
    print(
        "Backing off {wait:0.1f} seconds after {tries} tries "
        "calling function {target} with kwargs "
        "{kwargs}".format(**details),
    )


class GPT3(LM):
    """Wrapper around OpenAI's GPT API.

    Args:
        model (str, optional): OpenAI supported LLM model to use. Defaults to "gpt-3.5-turbo-instruct".
        api_key (Optional[str], optional): API provider Authentication token. use Defaults to None.
        api_provider (Literal["openai"], optional): The API provider to use. Defaults to "openai".
        model_type (Literal["chat", "text"], optional): The type of model that was specified. Mainly to decide the optimal prompting strategy. Defaults to "text".
        **kwargs: Additional arguments to pass to the API provider.
    """

    def __init__(
            self,
            model: str = "gpt-3.5-turbo-instruct",
            api_key: Optional[str] = None,
            api_provider: Literal["openai"] = "openai",
            api_base: Optional[str] = None,
            base_url: Optional[str] = None,
            model_type: Literal["chat", "text"] = None,
            system_prompt: Optional[str] = None,
            http_client: Optional[httpx.Client] = None,
            default_headers: Optional[dict[str, str]] = None,
            **kwargs,
    ):
        super().__init__(model)
        self.provider = "openai"
        openai.api_type = api_provider
        openai.default_headers = default_headers or {}

        self.system_prompt = system_prompt

        assert (
                api_provider != "azure"
        ), "Azure functionality with base OpenAI has been deprecated, please use dspy.AzureOpenAI instead."

        default_model_type = (
            "chat"
            if ("gpt-3.5" in model or "turbo" in model or "gpt-4" in model or "o1" in model) and ("instruct" not in model)
            else "text"
        )
        self.model_type = model_type if model_type else default_model_type

        if api_key:
            openai.api_key = api_key
        api_base = base_url or api_base
        if api_base:
            if OPENAI_LEGACY:
                openai.api_base = api_base
            else:
                openai.base_url = api_base
        if http_client:
            openai.http_client = http_client

        self.kwargs = {
            "temperature": 0.0,
            "max_tokens": 150,
            "top_p": 1,
            "frequency_penalty": 0,
            "presence_penalty": 0,
            "n": 1,
            **kwargs,
        }  # TODO: add kwargs above for </s>

        self.kwargs["model"] = model
        self.history: list[dict[str, Any]] = []

    def _openai_client(self):
        return openai

    def log_usage(self, response):
        """Log the total tokens from the OpenAI API response."""
        usage_data = response.get("usage")
        if usage_data:
            total_tokens = usage_data.get("total_tokens")
            logging.debug(f"OpenAI Response Token Usage: {total_tokens}")

    def basic_request(self, prompt: str, **kwargs):
        raw_kwargs = kwargs

        kwargs = {**self.kwargs, **kwargs}
        if "o1" in self.kwargs["model"]:
            if "max_tokens" in kwargs:
                max_tokens = kwargs.pop("max_tokens")
                kwargs["max_completion_tokens"] = max_tokens
            kwargs.pop("temperature")
            
        if self.model_type == "chat":
            # caching mechanism requires hashable kwargs
            messages = [{"role": "user", "content": prompt}]
            if self.system_prompt:
                messages.insert(0, {"role": "system", "content": self.system_prompt})
            kwargs["messages"] = messages
            kwargs = {"stringify_request": json.dumps(kwargs)}
            response = chat_request(**kwargs)

        else:
            kwargs["prompt"] = prompt
            response = completions_request(**kwargs)

        history = {
            "prompt": prompt,
            "response": response,
            "kwargs": kwargs,
            "raw_kwargs": raw_kwargs,
        }
        self.history.append(history)

        return response

    @backoff.on_exception(
        backoff.expo,
        ERRORS,
        max_time=settings.backoff_time,
        on_backoff=backoff_hdlr,
    )
    def request(self, prompt: str, **kwargs):
        """Handles retrieval of GPT-3 completions whilst handling rate limiting and caching."""
        if "model_type" in kwargs:
            del kwargs["model_type"]

        return self.basic_request(prompt, **kwargs)

    def _get_choice_text(self, choice: dict[str, Any]) -> str:
        if self.model_type == "chat":
            return choice["message"]["content"]
        return choice["text"]

    def __call__(
            self,
            prompt: str,
            only_completed: bool = True,
            return_sorted: bool = False,
            **kwargs,
    ) -> list[dict[str, Any]]:
        """Retrieves completions from GPT-3.

        Args:
            prompt (str): prompt to send to GPT-3
            only_completed (bool, optional): return only completed responses and ignores completion due to length. Defaults to True.
            return_sorted (bool, optional): sort the completion choices using the returned probabilities. Defaults to False.

        Returns:
            list[dict[str, Any]]: list of completion choices
        """

        assert only_completed, "for now"
        assert return_sorted is False, "for now"

        # if kwargs.get("n", 1) > 1:
        #     if self.model_type == "chat":
        #         kwargs = {**kwargs}
        #     else:
        #         kwargs = {**kwargs, "logprobs": 5}

        response = self.request(prompt, **kwargs)

        self.log_usage(response)
        choices = response["choices"]

        completed_choices = [c for c in choices if c["finish_reason"] != "length"]

        if only_completed and len(completed_choices):
            choices = completed_choices

        if kwargs.get("logprobs", False):
            completions = [{'text': self._get_choice_text(c), 'logprobs': c["logprobs"]} for c in choices]
        else:
            completions = [self._get_choice_text(c) for c in choices]

        if return_sorted and kwargs.get("n", 1) > 1:
            scored_completions = []

            for c in choices:
                tokens, logprobs = (
                    c["logprobs"]["tokens"],
                    c["logprobs"]["token_logprobs"],
                )

                if "<|endoftext|>" in tokens:
                    index = tokens.index("<|endoftext|>") + 1
                    tokens, logprobs = tokens[:index], logprobs[:index]

                avglog = sum(logprobs) / len(logprobs)
                scored_completions.append((avglog, self._get_choice_text(c), logprobs))
            scored_completions = sorted(scored_completions, reverse=True)
            if logprobs:
                completions = [{'text': c, 'logprobs': lp} for _, c, lp in scored_completions]
            else:
                completions = [c for _, c in scored_completions]

        return completions


@CacheMemory.cache
def cached_gpt3_request_v2(**kwargs):
    return openai.Completion.create(**kwargs)


@functools.lru_cache(maxsize=None if cache_turn_on else 0)
@NotebookCacheMemory.cache
def cached_gpt3_request_v2_wrapped(**kwargs):
    return cached_gpt3_request_v2(**kwargs)


@CacheMemory.cache
def _cached_gpt3_turbo_request_v2(**kwargs) -> OpenAIObject:
    if "stringify_request" in kwargs:
        kwargs = json.loads(kwargs["stringify_request"])
    return cast(OpenAIObject, openai.ChatCompletion.create(**kwargs))


@functools.lru_cache(maxsize=None if cache_turn_on else 0)
@NotebookCacheMemory.cache
def _cached_gpt3_turbo_request_v2_wrapped(**kwargs) -> OpenAIObject:
    return _cached_gpt3_turbo_request_v2(**kwargs)


@CacheMemory.cache
def v1_cached_gpt3_request_v2(**kwargs):
    return openai.completions.create(**kwargs)


@functools.lru_cache(maxsize=None if cache_turn_on else 0)
@NotebookCacheMemory.cache
def v1_cached_gpt3_request_v2_wrapped(**kwargs):
    return v1_cached_gpt3_request_v2(**kwargs)


@CacheMemory.cache
def v1_cached_gpt3_turbo_request_v2(**kwargs):
    if "stringify_request" in kwargs:
        kwargs = json.loads(kwargs["stringify_request"])
    return openai.chat.completions.create(**kwargs)


@functools.lru_cache(maxsize=None if cache_turn_on else 0)
@NotebookCacheMemory.cache
def v1_cached_gpt3_turbo_request_v2_wrapped(**kwargs):
    return v1_cached_gpt3_turbo_request_v2(**kwargs)


def chat_request(**kwargs):
    if OPENAI_LEGACY:
        return _cached_gpt3_turbo_request_v2_wrapped(**kwargs)

    return v1_cached_gpt3_turbo_request_v2_wrapped(**kwargs).model_dump()


def completions_request(**kwargs):
    if OPENAI_LEGACY:
        return cached_gpt3_request_v2_wrapped(**kwargs)

    return v1_cached_gpt3_request_v2_wrapped(**kwargs).model_dump()



import dspy

from .predict import Module

# TODO: FIXME: Insert this right before the *first* output field. Also rewrite this to use the new signature system.

class ChainOfThoughtWithHint(Module):
    def __init__(self, signature, rationale_type=None, **config):
        self.signature = dspy.ensure_signature(signature)
        self.module = dspy.ChainOfThought(signature, rationale_type=rationale_type, **config)
    
    def forward(self, **kwargs):
        if 'hint' in kwargs and kwargs['hint']:
            hint = f"\n\t\t(secret hint: {kwargs.pop('hint')})"
            original_kwargs = kwargs.copy()
            
            # Convert the first field's value to string and append the hint
            last_key = list(self.signature.input_fields.keys())[-1]
            kwargs[last_key] = str(kwargs[last_key]) + hint

            # Run CoT then update the trace with original kwargs, i.e. without the hint.
            pred = self.module(**kwargs)
            this_trace = dspy.settings.trace[-1]
            dspy.settings.trace[-1] = (this_trace[0], original_kwargs, this_trace[2])
            return pred
        
        return self.module(**kwargs)


"""
TODO: In principle, we can update the field's prefix during forward too to fill any thing based on the input args.

IF the user didn't overwrite our default rationale_type.
"""


import logging
import sys
import tqdm
import dspy
import signal
import threading
import traceback
import contextlib

from tqdm.contrib.logging import logging_redirect_tqdm
from concurrent.futures import ThreadPoolExecutor, as_completed


logger = logging.getLogger(__name__)


class ParallelExecutor:
    def __init__(
        self,
        num_threads,
        max_errors=5,
        disable_progress_bar=False,
        provide_traceback=False,
        compare_results=False,
    ):
        self.num_threads = num_threads
        self.disable_progress_bar = disable_progress_bar
        self.max_errors = max_errors
        self.provide_traceback = provide_traceback
        self.compare_results = compare_results

        self.error_count = 0
        self.error_lock = threading.Lock()
        self.cancel_jobs = threading.Event()


    def execute(self, function, data):
        wrapped_function = self._wrap_function(function)
        if self.num_threads == 1:
            return self._execute_single_thread(wrapped_function, data)
        else:
            return self._execute_multi_thread(wrapped_function, data)


    def _wrap_function(self, function):
        # Wrap the function with threading context and error handling
        def wrapped(item, parent_id=None):
            thread_stacks = dspy.settings.stack_by_thread
            current_thread_id = threading.get_ident()
            creating_new_thread = current_thread_id not in thread_stacks

            assert creating_new_thread or threading.get_ident() == dspy.settings.main_tid

            if creating_new_thread:
                # If we have a parent thread ID, copy its stack. TODO: Should the caller just pass a copy of the stack?
                if parent_id and parent_id in thread_stacks:
                    thread_stacks[current_thread_id] = list(thread_stacks[parent_id])
                else:
                    thread_stacks[current_thread_id] = list(dspy.settings.main_stack)

                # TODO: Consider the behavior below.
                # import copy; thread_stacks[current_thread_id].append(copy.deepcopy(thread_stacks[current_thread_id][-1]))

            try:
                return function(item)
            except Exception as e:
                with self.error_lock:
                    self.error_count += 1
                    current_error_count = self.error_count
                if current_error_count >= self.max_errors:
                    self.cancel_jobs.set()
                    raise e
                if self.provide_traceback:
                    logger.error(
                        f"Error processing item {item}: {e}\nStack trace:\n{traceback.format_exc()}"
                    )
                else:
                    logger.error(
                        f"Error processing item {item}: {e}. Set `provide_traceback=True` to see the stack trace."
                    )
                return None
            finally:
                if creating_new_thread:
                    del thread_stacks[threading.get_ident()]
        return wrapped


    def _execute_single_thread(self, function, data):
        results = []
        pbar = tqdm.tqdm(
            total=len(data),
            dynamic_ncols=True,
            disable=self.disable_progress_bar,
            file=sys.stdout,
        )
        for item in data:
            with logging_redirect_tqdm():
                if self.cancel_jobs.is_set():
                    break
                result = function(item)
                results.append(result)
                if self.compare_results:
                    # Assumes score is the last element of the result tuple
                    self._update_progress(pbar, sum([r[-1] for r in results if r is not None]), len([r for r in data if r is not None]))
                else:
                    self._update_progress(pbar, len(results), len(data))
        pbar.close()
        if self.cancel_jobs.is_set():
            logger.warning("Execution was cancelled due to errors.")
            raise Exception("Execution was cancelled due to errors.")
        return results


    def _update_progress(self, pbar, nresults, ntotal):
        if self.compare_results:
            pbar.set_description(f"Average Metric: {nresults:.2f} / {ntotal} ({round(100 * nresults / ntotal, 1) if ntotal > 0 else 0}%)")
        else:
            pbar.set_description(f"Processed {nresults} / {ntotal} examples")
        pbar.update()


    def _execute_multi_thread(self, function, data):
        results = [None] * len(data)  # Pre-allocate results list to maintain order
        job_cancelled = "cancelled"

        @contextlib.contextmanager
        def interrupt_handler_manager():
            """Sets the cancel_jobs event when a SIGINT is received, only in the main thread."""
            if threading.current_thread() is threading.main_thread():
                default_handler = signal.getsignal(signal.SIGINT)

                def interrupt_handler(sig, frame):
                    self.cancel_jobs.set()
                    logger.warning("Received SIGINT. Cancelling execution.")
                    default_handler(sig, frame)

                signal.signal(signal.SIGINT, interrupt_handler)
                try:
                    yield
                finally:
                    signal.signal(signal.SIGINT, default_handler)
            else:
                # If not in the main thread, skip setting signal handlers
                yield

        def cancellable_function(index_item, parent_id=None):
            index, item = index_item
            if self.cancel_jobs.is_set():
                return index, job_cancelled
            return index, function(item, parent_id)
        
        parent_id = threading.get_ident() if threading.current_thread() is not threading.main_thread() else None

        with ThreadPoolExecutor(max_workers=self.num_threads) as executor, interrupt_handler_manager():
            futures = {executor.submit(cancellable_function, pair, parent_id): pair for pair in enumerate(data)}
            pbar = tqdm.tqdm(
                total=len(data),
                dynamic_ncols=True,
                disable=self.disable_progress_bar,
                file=sys.stdout,
            )

            for future in as_completed(futures):
                index, result = future.result()
            
                if result is job_cancelled:
                    continue
                results[index] = result

                if self.compare_results:
                    # Assumes score is the last element of the result tuple
                    self._update_progress(pbar, sum([r[-1] for r in results if r is not None]), len([r for r in results if r is not None]))
                else:
                    self._update_progress(pbar, len([r for r in results if r is not None]), len(data))
            pbar.close()
        if self.cancel_jobs.is_set():
            logger.warning("Execution was cancelled due to errors.")
            raise Exception("Execution was cancelled due to errors.")
        return results



import litellm
import numpy as np


class Embedding:
    """DSPy embedding class.

    The class for computing embeddings for text inputs. This class provides a unified interface for both:

    1. Hosted embedding models (e.g. OpenAI's text-embedding-3-small) via litellm integration
    2. Custom embedding functions that you provide

    For hosted models, simply pass the model name as a string (e.g. "openai/text-embedding-3-small"). The class will use
    litellm to handle the API calls and caching.

    For custom embedding models, pass a callable function that:
    - Takes a list of strings as input.
    - Returns embeddings as either:
        - A 2D numpy array of float32 values
        - A 2D list of float32 values
    - Each row should represent one embedding vector

    Args:
        model: The embedding model to use. This can be either a string (representing the name of the hosted embedding
            model, must be an embedding model supported by litellm) or a callable that represents a custom embedding
            model.

    Examples:
        Example 1: Using a hosted model.

        ```python
        import dspy

        embedder = dspy.Embedding("openai/text-embedding-3-small")
        embeddings = embedder(["hello", "world"])

        assert embeddings.shape == (2, 1536)
        ```

        Example 2: Using a custom function.

        ```python
        import dspy

        def my_embedder(texts):
            return np.random.rand(len(texts), 10)

        embedder = dspy.Embedding(my_embedder)
        embeddings = embedder(["hello", "world"])

        assert embeddings.shape == (2, 10)
        ```
    """

    def __init__(self, model):
        self.model = model

    def __call__(self, inputs, caching=True, **kwargs):
        """Compute embeddings for the given inputs.

        Args:
            inputs: The inputs to compute embeddings for, can be a single string or a list of strings.
            caching: Whether to cache the embedding response, only valid when using a hosted embedding model.
            kwargs: Additional keyword arguments to pass to the embedding model.

        Returns:
            A 2-D numpy array of embeddings, one embedding per row.
        """
        if isinstance(inputs, str):
            inputs = [inputs]
        if isinstance(self.model, str):
            embedding_response = litellm.embedding(model=self.model, input=inputs, caching=caching, **kwargs)
            return np.array([data["embedding"] for data in embedding_response.data], dtype=np.float32)
        elif callable(self.model):
            return np.array(self.model(inputs, **kwargs), dtype=np.float32)
        else:
            raise ValueError(f"`model` in `dspy.Embedding` must be a string or a callable, but got {type(self.model)}.")



import os
from collections.abc import Iterable
from typing import Any, Optional

import backoff

from dsp.modules.lm import LM
from dsp.utils.settings import settings

try:
    import google.generativeai as genai
    from google.api_core.exceptions import GoogleAPICallError
    google_api_error = GoogleAPICallError
except ImportError:
    google_api_error = Exception
    # print("Not loading Google because it is not installed.")


def backoff_hdlr(details):
    """Handler from https://pypi.org/project/backoff/"""
    print(
        "Backing off {wait:0.1f} seconds after {tries} tries "
        "calling function {target} with kwargs "
        "{kwargs}".format(**details),
    )


def giveup_hdlr(details):
    """wrapper function that decides when to give up on retry"""
    if "rate limits" in details.message:
        return False
    return True


BLOCK_ONLY_HIGH = [
  {
    "category": "HARM_CATEGORY_HARASSMENT",
    "threshold": "BLOCK_ONLY_HIGH",
  },
  {
    "category": "HARM_CATEGORY_HATE_SPEECH",
    "threshold": "BLOCK_ONLY_HIGH",
  },
  {
    "category": "HARM_CATEGORY_SEXUALLY_EXPLICIT",
    "threshold": "BLOCK_ONLY_HIGH",
  },
  {
    "category": "HARM_CATEGORY_DANGEROUS_CONTENT",
    "threshold": "BLOCK_ONLY_HIGH",
  },
]


class Google(LM):
    """Wrapper around Google's API.

    Currently supported models include `gemini-pro-1.0`.
    """

    def __init__(
        self,
        model: str = "models/gemini-1.0-pro",
        api_key: Optional[str] = None,
        safety_settings: Optional[Iterable] = BLOCK_ONLY_HIGH,
        **kwargs,
    ):
        """
        Parameters
        ----------
        model : str
            Which pre-trained model from Google to use?
            Choices are [`gemini-pro-1.0`]
        api_key : str
            The API key for Google.
            It can be obtained from https://cloud.google.com/generative-ai-studio
        **kwargs: dict
            Additional arguments to pass to the API provider.
        """
        super().__init__(model)
        api_key = os.environ.get("GOOGLE_API_KEY") if api_key is None else api_key
        genai.configure(api_key=api_key)

        # Google API uses "candidate_count" instead of "n" or "num_generations"
        # For now, google API only supports 1 generation at a time. Raises an error if candidate_count > 1
        num_generations = kwargs.pop("n", kwargs.pop("num_generations", 1))

        self.provider = "google"
        kwargs = {
            "candidate_count": 1,
            "temperature": 0.0 if "temperature" not in kwargs else kwargs["temperature"],
            "max_output_tokens": 2048,
            "top_p": 1,
            "top_k": 1,
            **kwargs,
        }

        self.config = genai.GenerationConfig(**kwargs)
        self.llm = genai.GenerativeModel(model_name=model,
                                         generation_config=self.config,
                                         safety_settings=safety_settings)

        self.kwargs = {
            "n": num_generations,
            **kwargs,
        }

        self.history: list[dict[str, Any]] = []

    def basic_request(self, prompt: str, **kwargs):
        raw_kwargs = kwargs
        kwargs = {
            **self.kwargs,
            **kwargs,
        }

        # Google disallows "n" arguments
        n = kwargs.pop("n", None)
        if n is not None and n > 1 and kwargs['temperature'] == 0.0:
            kwargs['temperature'] = 0.7

        response = self.llm.generate_content(prompt, generation_config=kwargs)

        history = {
            "prompt": prompt,
            "response": [response],
            "kwargs": kwargs,
            "raw_kwargs": raw_kwargs,
        }
        self.history.append(history)

        return response

    @backoff.on_exception(
        backoff.expo,
        (google_api_error),
        max_time=settings.backoff_time,
        max_tries=8,
        on_backoff=backoff_hdlr,
        giveup=giveup_hdlr,
    )
    def request(self, prompt: str, **kwargs):
        """Handles retrieval of completions from Google whilst handling API errors"""
        return self.basic_request(prompt, **kwargs)

    def __call__(
        self,
        prompt: str,
        only_completed: bool = True,
        return_sorted: bool = False,
        **kwargs,
    ):
        assert only_completed, "for now"
        assert return_sorted is False, "for now"

        n = kwargs.pop("n", 1)

        completions = []
        for i in range(n):
            response = self.request(prompt, **kwargs)
            completions.append(response.parts[0].text)

        return completions



from typing import Optional, Union


def passages2text(passages: Union[str, list, tuple]) -> str:
    """Formats the given one or more passages into a single structured string."""
    if isinstance(passages, str):
        return passages

    assert type(passages) in [list, tuple]

    if len(passages) == 0:
        return "N/A"

    if len(passages) == 1:
        return f"«{passages[0]}»"

    return "\n".join([f"[{idx+1}] «{txt}»" for idx, txt in enumerate(passages)])


# def passages2textV2(passages: Union[str, list, tuple]) -> str:
#     """Formats the given one or more passages into a single structured string."""
#     if isinstance(passages, str):
#         return passages

#     assert type(passages) in [list, tuple]

#     def psg2text(psg):
#         try:
#             title, snippet = psg.split("|", 1)
#             return f"Title: {title.strip()} | Snippet: «{snippet.strip()}»"
#         except Exception:
#             pass
        
#         return f"«{psg}»"

#     if len(passages) == 0:
#         return "N/A"

#     if len(passages) == 1:
#         return psg2text(passages[0])

#     return "\n".join([f"[{idx+1}] {psg2text(txt)}" for idx, txt in enumerate(passages)])


def format_answers(answers: Union[str, list]) -> Optional[str]:
    """Parses the given answers and returns the appropriate answer string.

    Args:
        answers (Union[str, list]): The answers to parse.

    Raises:
        ValueError: when instance is of type list and has no answers
        ValueError: when is not of type list or str

    Returns:
        _type_: Optional[str]
    """
    if isinstance(answers, list):
        if len(answers) >= 1:
            return str(answers[0]).strip()
        if len(answers) == 0:
            raise ValueError("No answers found")
    elif isinstance(answers, str):
        return answers
    else:
        raise ValueError(f"Unable to parse answers of type {type(answers)}")



import logging
import random
from functools import lru_cache

from pydantic import BaseModel

import dsp
from dspy.predict.parameter import Parameter
from dspy.primitives.prediction import Prediction
from dspy.primitives.program import Module
from dspy.signatures.signature import ensure_signature, signature_to_template
from dspy.utils.callback import with_callbacks
from dspy.adapters.image_utils import Image

@lru_cache(maxsize=None)
def warn_once(msg: str):
    logging.warning(msg)

class Predict(Module, Parameter):
    def __init__(self, signature, _parse_values=True, callbacks=None, **config):
        self.stage = random.randbytes(8).hex()
        self.signature = ensure_signature(signature)
        self.config = config
        self.callbacks = callbacks or []
        self._parse_values = _parse_values
        self.reset()

    def reset(self):
        self.lm = None
        self.traces = []
        self.train = []
        self.demos = []

    def dump_state(self, save_verbose=None):
        if save_verbose:
            logging.warning(
                "`save_verbose` is deprecated and will be removed in DSPy 2.6.0 release. Currently `save_verbose` "
                "does nothing."
            )

        state_keys = ["lm", "traces", "train"]
        state = {k: getattr(self, k) for k in state_keys}

        state["demos"] = []
        for demo in self.demos:
            demo = demo.copy()

            for field in demo:
                # FIXME: Saving BaseModels as strings in examples doesn't matter because you never re-access as an object
                # It does matter for images
                if isinstance(demo[field], Image):
                    demo[field] = demo[field].model_dump()
                elif isinstance(demo[field], BaseModel):
                    demo[field] = demo[field].model_dump_json()

            state["demos"].append(demo)

        state["signature"] = self.signature.dump_state()
        # `extended_signature` is a special field for `Predict`s like CoT.
        if hasattr(self, "extended_signature"):
            state["extended_signature"] = self.extended_signature.dump_state()
        return state

    def load_state(self, state, use_legacy_loading=False):
        """Load the saved state of a `Predict` object.

        Args:
            state (dict): The saved state of a `Predict` object.
            use_legacy_loading (bool): Whether to use the legacy loading method. Only use it when you are loading a
                saved state from a version of DSPy prior to v2.5.3.
        Returns:
            self: Returns self to allow method chaining
        """
        if use_legacy_loading:
            self._load_state_legacy(state)
            return self
            
        if "signature" not in state:
            # Check if the state is from a version of DSPy prior to v2.5.3.
            raise ValueError(
                "The saved state is from a version of DSPy prior to v2.5.3. Please use `use_legacy_loading=True` to "
                "load the state."
            )

        excluded_keys = ["signature", "extended_signature"]
        for name, value in state.items():
            # `excluded_keys` are fields that go through special handling.
            if name not in excluded_keys:
                setattr(self, name, value)
        
        # FIXME: Images are getting special treatment, but all basemodels initialized from json should be converted back to objects
        for demo in self.demos:
            for field in demo:
                if isinstance(demo[field], dict) and "url" in demo[field]:
                    url = demo[field]["url"]
                    if not isinstance(url, str):
                        raise ValueError(f"Image URL must be a string, got {type(url)}")
                    demo[field] = Image(url=url)
                    
        self.signature = self.signature.load_state(state["signature"])

        if "extended_signature" in state:
            self.extended_signature = self.extended_signature.load_state(state["extended_signature"])

        return self

    def _load_state_legacy(self, state):
        """Legacy state loading for backwards compatibility.

        This method is used to load the saved state of a `Predict` object from a version of DSPy prior to v2.5.3.
        Returns:
            self: Returns self to allow method chaining
        """
        for name, value in state.items():
            setattr(self, name, value)

        # Reconstruct the signature.
        if "signature_instructions" in state:
            instructions = state["signature_instructions"]
            self.signature = self.signature.with_instructions(instructions)

        if "signature_prefix" in state:
            prefix = state["signature_prefix"]
            *_, last_key = self.signature.fields.keys()
            self.signature = self.signature.with_updated_fields(last_key, prefix=prefix)

        # Some special stuff for CoT.
        if "extended_signature_instructions" in state:
            instructions = state["extended_signature_instructions"]
            self.extended_signature = self.extended_signature.with_instructions(instructions)

        if "extended_signature_prefix" in state:
            prefix = state["extended_signature_prefix"]
            *_, last_key = self.extended_signature.fields.keys()
            self.extended_signature = self.extended_signature.with_updated_fields(last_key, prefix=prefix)

        return self

    def load(self, path, return_self=False):
        """Load a saved state from a file.
        
        Args:
            path (str): Path to the saved state file
            return_self (bool): If True, returns self to allow method chaining. Default is False for backwards compatibility.
        
        Returns:
            Union[None, Predict]: Returns None if return_self is False (default), returns self if return_self is True
        """
        super().load(path)
        return self if return_self else None

    @with_callbacks
    def __call__(self, **kwargs):
        return self.forward(**kwargs)

    def forward(self, **kwargs):
        assert not dsp.settings.compiling, "It's no longer ever the case that .compiling is True"

        # Extract the three privileged keyword arguments.
        new_signature = ensure_signature(kwargs.pop("new_signature", None))
        signature = ensure_signature(kwargs.pop("signature", self.signature))
        demos = kwargs.pop("demos", self.demos)
        config = dict(**self.config, **kwargs.pop("config", {}))

        # Get the right LM to use.
        lm = kwargs.pop("lm", self.lm) or dsp.settings.lm
        assert lm is not None, "No LM is loaded."

        # If temperature is 0.0 but its n > 1, set temperature to 0.7.
        temperature = config.get("temperature")
        temperature = lm.kwargs["temperature"] if temperature is None else temperature
        num_generations = config.get("n") or lm.kwargs.get("n") or lm.kwargs.get("num_generations") or 1

        if (temperature is None or temperature <= 0.15) and num_generations > 1:
            config["temperature"] = 0.7

        if new_signature is not None:
            signature = new_signature

        if not all(k in kwargs for k in signature.input_fields):
            present = [k for k in signature.input_fields if k in kwargs]
            missing = [k for k in signature.input_fields if k not in kwargs]
            print(f"WARNING: Not all input fields were provided to module. Present: {present}. Missing: {missing}.")

        import dspy

        if isinstance(lm, dspy.LM):
            completions = v2_5_generate(lm, config, signature, demos, kwargs, _parse_values=self._parse_values)
        else:
            warn_once(
                "\t*** In DSPy 2.5, all LM clients except `dspy.LM` are deprecated, "
                "underperform, and are about to be deleted. ***\n"
                f" \t\tYou are using the client {lm.__class__.__name__}, which will be removed in DSPy 2.6.\n"
                " \t\tChanging the client is straightforward and will let you use new features (Adapters) that"
                " improve the consistency of LM outputs, especially when using chat LMs. \n\n"
                " \t\tLearn more about the changes and how to migrate at\n"
                " \t\thttps://github.com/stanfordnlp/dspy/blob/main/examples/migration.ipynb"
            )

            if dsp.settings.experimental:
                completions = new_generate(lm, signature, dsp.Example(demos=demos, **kwargs), **config)
            else:
                completions = old_generate(demos, signature, kwargs, config, self.lm, self.stage)

        pred = Prediction.from_completions(completions, signature=signature)

        if kwargs.pop("_trace", True) and dsp.settings.trace is not None:
            trace = dsp.settings.trace
            trace.append((self, {**kwargs}, pred))

        return pred

    def update_config(self, **kwargs):
        self.config = {**self.config, **kwargs}

    def get_config(self):
        return self.config

    def __repr__(self):
        return f"{self.__class__.__name__}({self.signature})"


def old_generate(demos, signature, kwargs, config, lm, stage):
    # Switch to legacy format for dsp.generate
    x = dsp.Example(demos=demos, **kwargs)
    template = signature_to_template(signature)

    if lm is None:
        x, C = dsp.generate(template, **config)(x, stage=stage)
    else:
        # Note: query_only=True means the instructions and examples are not included.
        with dsp.settings.context(lm=lm, query_only=True):
            x, C = dsp.generate(template, **config)(x, stage=stage)

    completions = []

    for c in C:
        completions.append({})
        for field in template.fields:
            if field.output_variable not in kwargs.keys():
                completions[-1][field.output_variable] = getattr(c, field.output_variable)

    return completions


def new_generate(lm, signature, example, max_depth=6, **kwargs):
    kwargs["stop"] = tuple(kwargs.get("stop", [])) or ("\n---",)

    # Generate and extract the fields.
    template = signature_to_template(signature, adapter=dsp.ExperimentalAdapter)
    prompt = template(example)
    completions = lm(prompt, **kwargs)
    completions = [template.extract(example, p) for p in completions]

    assert all(set(signature.input_fields).issubset(set(c.keys())) for c in completions), "Missing input keys."

    # Find the completions that are most complete.
    field_names = [field.input_variable for field in template.fields]
    for field_idx, key in enumerate(field_names):
        completions_ = [c for c in completions if key in c.keys() and c[key] is not None]
        completions = completions_ or completions
        if len(completions_) == 0:
            break

    # If none of the completions is completed (i.e., none has the final field set).
    if len(completions_) == 0:
        # Pick the first completion that has gone farthest.
        completion = completions[0]

        for field_idx_ in range(field_idx + 1, len(field_names)):
            if field_names[field_idx_] in completion:
                del completion[field_names[field_idx_]]

        # Recurse with greedy decoding.
        new_kwargs = {
            **kwargs,
            "n": 1,
            "temperature": 0.0,
        }

        assert max_depth > 0
        return new_generate(lm, signature, completion, max_depth=max_depth - 1, **new_kwargs)

    # Keep only output fields.
    completions = [{k: v for k, v in c.items() if k in signature.output_fields} for c in completions]

    return completions


def v2_5_generate(lm, lm_kwargs, signature, demos, inputs, _parse_values=True):
    import dspy

    adapter = dspy.settings.adapter or dspy.ChatAdapter()

    return adapter(
        lm, lm_kwargs=lm_kwargs, signature=signature, demos=demos, inputs=inputs, _parse_values=_parse_values
    )

# TODO: get some defaults during init from the context window?
# # TODO: FIXME: Hmm, I guess expected behavior is that contexts can
# affect execution. Well, we need to determine whether context dominates, __init__ demoninates, or forward dominates.
# Generally, unless overwritten, we'd see n=None, temperature=None.
# That will eventually mean we have to learn them.



from .aggregation import majority
from .chain_of_thought import ChainOfThought
from .chain_of_thought_with_hint import ChainOfThoughtWithHint
from .knn import KNN
from .multi_chain_comparison import MultiChainComparison
from .predict import Predict
from .program_of_thought import ProgramOfThought
from .react import ReAct
from .retry import Retry
from .parallel import Parallel


import re
import string
import unicodedata
from collections import Counter

from dsp.utils.utils import print_message


def EM(prediction, answers_list):
    assert type(answers_list) == list

    return max(em_score(prediction, ans) for ans in answers_list)


def F1(prediction, answers_list):
    assert type(answers_list) == list

    return max(f1_score(prediction, ans) for ans in answers_list)


def HotPotF1(prediction, answers_list):
    assert type(answers_list) == list

    return max(hotpot_f1_score(prediction, ans) for ans in answers_list)


def nF1(history, prediction, answers_list, return_recall=False):
    assert type(answers_list) == list

    return max(novel_f1_score(history, prediction, ans, return_recall=return_recall) for ans in answers_list)


def normalize_text(s):
    s = unicodedata.normalize('NFD', s)

    def remove_articles(text):
        return re.sub(r'\b(a|an|the)\b', ' ', text)

    def white_space_fix(text):
        return ' '.join(text.split())

    def remove_punc(text):
        exclude = set(string.punctuation)
        return ''.join(ch for ch in text if ch not in exclude)

    def lower(text):
        return text.lower()

    return white_space_fix(remove_articles(remove_punc(lower(s))))


def em_score(prediction, ground_truth):
    return normalize_text(prediction) == normalize_text(ground_truth)


# See: https://github.com/hotpotqa/hotpot/blob/master/hotpot_evaluate_v1.py
# See: https://rajpurkar.github.io/SQuAD-explorer/ under Evaluation Script
# See: QReCC's

def f1_score(prediction, ground_truth):
    prediction_tokens = normalize_text(prediction).split()
    ground_truth_tokens = normalize_text(ground_truth).split()

    common = Counter(prediction_tokens) & Counter(ground_truth_tokens)
    num_same = sum(common.values())

    if len(prediction_tokens) == len(ground_truth_tokens) == 0:
        # Unlike most tasks, QReCC and SQuAD-2.0 assign 1.0 in this edge case. We don't for uniformity.
        print_message(
            "\n#> F1 Metric: Rare edge case of len(prediction_tokens) == len(ground_truth_tokens) == 0.\n")

    if num_same == 0:
        return 0

    precision = 1.0 * num_same / len(prediction_tokens)
    recall = 1.0 * num_same / len(ground_truth_tokens)
    f1 = (2 * precision * recall) / (precision + recall)

    return f1


def hotpot_f1_score(prediction, ground_truth):
    normalized_prediction = normalize_text(prediction)
    normalized_ground_truth = normalize_text(ground_truth)

    if normalized_prediction in ['yes', 'no', 'noanswer'] and normalized_prediction != normalized_ground_truth:
        return 0
    if normalized_ground_truth in ['yes', 'no', 'noanswer'] and normalized_prediction != normalized_ground_truth:
        return 0

    prediction_tokens = normalized_prediction.split()
    ground_truth_tokens = normalized_ground_truth.split()
    common = Counter(prediction_tokens) & Counter(ground_truth_tokens)
    num_same = sum(common.values())
    if num_same == 0:
        return 0
    precision = 1.0 * num_same / len(prediction_tokens)
    recall = 1.0 * num_same / len(ground_truth_tokens)
    f1 = (2 * precision * recall) / (precision + recall)
    return f1


def precision_score(prediction, ground_truth):
    prediction_tokens = normalize_text(prediction).split()
    ground_truth_tokens = normalize_text(ground_truth).split()

    common = Counter(prediction_tokens) & Counter(ground_truth_tokens)
    num_same = sum(common.values())

    if len(prediction_tokens) == len(ground_truth_tokens) == 0:
        # Unlike most tasks, QReCC and SQuAD-2.0 assign 1.0 in this edge case. We don't for uniformity.
        print_message(
            "\n#> F1 Metric: Rare edge case of len(prediction_tokens) == len(ground_truth_tokens) == 0.\n")

    if num_same == 0:
        return 0

    precision = 1.0 * num_same / len(prediction_tokens)

    return precision


# Source: https://gist.github.com/sebleier/554280
stopwords = ["i", "me", "my", "myself", "we", "our", "ours", "ourselves", "you", "your", "yours", "yourself",
             "yourselves", "he", "him", "his", "himself", "she", "her", "hers", "herself", "it", "its", "itself",
             "they", "them", "their", "theirs", "themselves", "what", "which", "who", "whom", "this", "that", "these",
             "those", "am", "is", "are", "was", "were", "be", "been", "being", "have", "has", "had", "having", "do",
             "does", "did", "doing", "a", "an", "the", "and", "but", "if", "or", "because", "as", "until", "while",
             "of", "at", "by", "for", "with", "about", "against", "between", "into", "through", "during", "before",
             "after", "above", "below", "to", "from", "up", "down", "in", "out", "on", "off", "over", "under", "again",
             "further", "then", "once", "here", "there", "when", "where", "why", "how", "all", "any", "both", "each",
             "few", "more", "most", "other", "some", "such", "no", "nor", "not", "only", "own", "same", "so", "than",
             "too", "very", "s", "t", "can", "will", "just", "don", "should", "now"]


def novel_f1_score(history, prediction, ground_truth, return_recall=False):
    history_tokens = normalize_text(history).split()
    prediction_tokens = normalize_text(prediction).split()
    ground_truth_tokens = normalize_text(ground_truth).split()

    history_tokens = set(history_tokens + stopwords)

    prediction_tokens = [
        t for t in prediction_tokens if t not in history_tokens]
    ground_truth_tokens = [
        t for t in ground_truth_tokens if t not in history_tokens]

    common = Counter(prediction_tokens) & Counter(ground_truth_tokens)
    num_same = sum(common.values())
    if num_same == 0:
        return 0

    precision = 1.0 * num_same / len(prediction_tokens)
    recall = 1.0 * num_same / len(ground_truth_tokens)
    f1 = (2 * precision * recall) / (precision + recall)

    if return_recall:
        return recall

    return f1



import random

import dspy
from dspy.propose.dataset_summary_generator import create_dataset_summary
from dspy.propose.utils import create_example_string, create_predictor_level_history_string, strip_prefix, get_dspy_source_code
from dspy.teleprompt.utils import get_signature, get_prompt_model

from .propose_base import Proposer

# Hardcoded variables (TODO: update)
MAX_INSTRUCT_IN_HISTORY = 5  # 10

TIPS = {
        "none": "",
        "creative": "Don't be afraid to be creative when creating the new instruction!",
        "simple": "Keep the instruction clear and concise.",
        "description": "Make sure your instruction is very informative and descriptive.",
        "high_stakes": "The instruction should include a high stakes scenario in which the LM must solve the task!",
        "persona": 'Include a persona that is relevant to the task in the instruction (ie. "You are a ...")',
    }

### SIGNATURES USED TO HELP WITH INSTRUCTION GENERATION ###

class DescribeProgram(dspy.Signature):
    (
        """Below is some pseudo-code for a pipeline that solves tasks with calls to language models. Please describe what type of task this program appears to be designed to solve, and how it appears to work."""
    )
    program_code = dspy.InputField(
        format=str,
        desc="Pseudocode for a language model program designed to solve a particular task.",
        prefix="PROGRAM CODE:",
    )
    program_example = dspy.InputField(
        format=str,
        desc="An example of the program in use.",
        prefix="EXAMPLE OF PROGRAM IN USE:",
    )
    program_description = dspy.OutputField(
        desc="Describe what task the program is designed to solve, and how it goes about solving this task.",
        prefix="SUMMARY OF PROGRAM ABOVE:",
    )


class DescribeModule(dspy.Signature):
    (
        """Below is some pseudo-code for a pipeline that solves tasks with calls to language models. Please describe the purpose of one of the specified module in this pipeline."""
    )
    program_code = dspy.InputField(
        format=str,
        desc="Pseudocode for a language model program designed to solve a particular task.",
        prefix="PROGRAM CODE:",
    )
    program_example = dspy.InputField(
        format=str,
        desc="An example of the program in use.",
        prefix="EXAMPLE OF PROGRAM IN USE:",
    )
    program_description = dspy.InputField(
        desc="Summary of the task the program is designed to solve, and how it goes about solving it.",
        prefix="SUMMARY OF PROGRAM ABOVE:",
    )
    module = dspy.InputField(
        desc="The module in the program that we want to describe.", prefix="MODULE:",
    )
    module_description = dspy.OutputField(
        desc="Description of the module's role in the broader program.",
        prefix="MODULE DESCRIPTION:",
    )


def generate_instruction_class(
    use_dataset_summary=True,
    program_aware=True,
    use_task_demos=True,
    use_instruct_history=True,
    use_tip=True,
):
    class GenerateSingleModuleInstruction(dspy.Signature):
        (
            """Use the information below to learn about a task that we are trying to solve using calls to an LM, then generate a new instruction that will be used to prompt a Language Model to better solve the task."""
        )
        if use_dataset_summary:
            dataset_description = dspy.InputField(
                desc="A description of the dataset that we are using.",
                prefix="DATASET SUMMARY:",
            )
        if program_aware:
            program_code = dspy.InputField(
                format=str,
                desc="Language model program designed to solve a particular task.",
                prefix="PROGRAM CODE:",
            )
            program_description = dspy.InputField(
                desc="Summary of the task the program is designed to solve, and how it goes about solving it.",
                prefix="PROGRAM DESCRIPTION:",
            )
            module = dspy.InputField(
                desc="The module to create an instruction for.", prefix="MODULE:",
            )
        task_demos = dspy.InputField(
            format=str,
            desc="Example inputs/outputs of our module.",
            prefix="TASK DEMO(S):",
        )
        if use_instruct_history:
            previous_instructions = dspy.InputField(
                format=str,
                desc="Previous instructions we've attempted, along with their associated scores.",
                prefix="PREVIOUS INSTRUCTIONS:",
            )
        basic_instruction = dspy.InputField(
            format=str, desc="Basic instruction.", prefix="BASIC INSTRUCTION:",
        )
        if use_tip:
            tip = dspy.InputField(
                format=str,
                desc="A suggestion for how to go about generating the new instruction.",
                prefix="TIP:",
            )
        proposed_instruction = dspy.OutputField(
            desc="Propose an instruction that will be used to prompt a Language Model to perform this task.",
            prefix="PROPOSED INSTRUCTION:",
        )

    return dspy.Predict(GenerateSingleModuleInstruction)

### CLASS RESPONSIBLE FOR GENERATING A NEW INSTRUCTION, USING THE HELPER SIGNATURES ABOVE ###

class GenerateModuleInstruction(dspy.Module):
    def __init__(
        self,
        program_code_string=None,
        use_dataset_summary=True,
        program_aware=False,
        use_task_demos=True,
        use_instruct_history=True,
        use_tip=True,
        verbose=False,
    ):
        super().__init__()
        self.use_dataset_summary = use_dataset_summary
        self.program_aware = program_aware
        self.use_task_demos = use_task_demos
        self.use_instruct_history = use_instruct_history
        self.use_tip = use_tip
        self.verbose = verbose

        self.program_code_string = program_code_string
        self.describe_program = dspy.Predict(DescribeProgram)
        self.describe_module = dspy.Predict(DescribeModule)
        self.generate_module_instruction = generate_instruction_class(
            use_dataset_summary=use_dataset_summary,
            program_aware=program_aware,
            use_task_demos=use_task_demos,
            use_instruct_history=use_instruct_history,
            use_tip=use_tip,
        )

    def forward(
        self,
        demo_candidates,
        pred_i,
        demo_set_i,
        program,
        previous_instructions,
        data_summary,
        max_demos=3,
        tip=None,
    ):
        # Construct full program demo or single module demo depending on whether or not we're using the full program
        task_demos = ""
        basic_instruction = get_signature(program.predictors()[pred_i]).instructions
        curr_demos_num = 0
        
        if self.use_task_demos:
            for example in demo_candidates[pred_i][demo_set_i]:
                if "augmented" in example.keys():
                    fields_to_use = get_signature(program.predictors()[pred_i]).fields
                    example_string = create_example_string(fields_to_use, example)
                    task_demos += f"{example_string}\n"
                    curr_demos_num += 1
                    if curr_demos_num >= max_demos:
                        break
        else:
            task_demos = "No task demos provided."

        # Summarize the program
        program_description = "Not available"
        module_code = "Not provided"
        module_description = "Not provided"
        if self.program_aware:
            try:
                program_description = strip_prefix(
                    self.describe_program(
                        program_code=self.program_code_string, program_example=task_demos,
                    ).program_description,
                )
                if self.verbose: print(f"PROGRAM DESCRIPTION: {program_description}")

                inputs = []
                outputs = []
                for field_name, field in get_signature(program.predictors()[pred_i]).fields.items():
                    # Access the '__dspy_field_type' from the extra metadata
                    dspy_field_type = field.json_schema_extra.get('__dspy_field_type')
                    
                    # Based on the '__dspy_field_type', append to the respective list
                    if dspy_field_type == "input":
                        inputs.append(field_name)
                    else:
                        outputs.append(field_name)

                module_code = f"{program.predictors()[pred_i].__class__.__name__}({', '.join(inputs)}) -> {', '.join(outputs)}"

                module_description = self.describe_module(
                    program_code=self.program_code_string,
                    program_description=program_description,
                    program_example=task_demos,
                    module=module_code,
                    max_depth=10,
                ).module_description
            except:
                if self.verbose: print("Error getting program description. Running without program aware proposer.")
                self.program_aware = False

        # Generate an instruction for our chosen module
        if self.verbose: print(f"task_demos {task_demos}")
        instruct = self.generate_module_instruction(
            dataset_description=data_summary,
            program_code=self.program_code_string,
            program_description=program_description,
            module=module_code,
            task_demos=task_demos,
            tip=tip,
            basic_instruction=basic_instruction,
            previous_instructions=previous_instructions,
            module_description=module_description,
        )
        if hasattr(instruct, "module_description"):
            module_description = strip_prefix(instruct.module_description)
            if self.verbose: print(f"MODULE DESCRIPTION: {module_description}")
        proposed_instruction = strip_prefix(instruct.proposed_instruction)

        return dspy.Prediction(proposed_instruction=proposed_instruction)

### CLASS USED TO GENERATE THE FULL SET OF INSTRUCTIONS GIVEN THE SPECIFIED CRITERIA ###

class GroundedProposer(Proposer):
    def __init__(
        self,
        prompt_model,
        program,
        trainset,
        view_data_batch_size=10,
        use_dataset_summary=True,
        program_aware=True,
        use_task_demos=True,
        use_instruct_history=True,
        use_tip=True,
        set_tip_randomly=True,
        set_history_randomly=True,
        verbose=False,
        rng=None
    ):
        super().__init__()
        self.program_aware = program_aware
        self.use_dataset_summary = use_dataset_summary
        self.use_task_demos = use_task_demos
        self.use_instruct_history = use_instruct_history
        self.use_tip = use_tip
        self.set_tip_randomly=set_tip_randomly
        self.set_history_randomly=set_history_randomly
        self.verbose = verbose
        self.rng = rng or random

        self.prompt_model = get_prompt_model(prompt_model)

        self.program_code_string = None
        if self.program_aware:
            try:
                self.program_code_string = get_dspy_source_code(program)
                if self.verbose: print("SOURCE CODE:",self.program_code_string)
            except Exception as e:
                print(f"Error getting source code: {e}.\n\nRunning without program aware proposer.")
                self.program_aware = False

        self.data_summary  = None
        if self.use_dataset_summary:
            try:
                self.data_summary = create_dataset_summary(
                    trainset=trainset, view_data_batch_size=view_data_batch_size, prompt_model=prompt_model,
                )
                if self.verbose: print(f"DATA SUMMARY: {self.data_summary}")
            except Exception as e:
                print(f"Error getting data summary: {e}.\n\nRunning without data aware proposer.")
                self.use_dataset_summary = False
                print("")

    def propose_instructions_for_program(
        self,
        trainset,
        program,
        demo_candidates,
        trial_logs,
        N,
        T,
        tip=None,
    ):
        """This method is responsible for returning the full set of new instructions for our program, given the specified criteria."""

        proposed_instructions = {}      

        if self.set_history_randomly:
            # Randomly select whether or not we're using instruction history
            use_history = self.rng.random() < 0.5
            self.use_instruct_history = use_history
            if self.verbose: print(f"Use history T/F: {self.use_instruct_history}")

        num_demos = max(len(demo_candidates[0]) if demo_candidates else N, 1)

        if not demo_candidates:
            if self.verbose: print("No demo candidates provided. Running without task demos.")
            self.use_task_demos = False

        # Create an instruction for each predictor 
        for pred_i, predictor in enumerate(program.predictors()):
            for demo_set_i in range(num_demos):
                if pred_i not in proposed_instructions:
                    proposed_instructions[pred_i] = []
                if self.set_tip_randomly:
                    if self.verbose: print("Using a randomly generated configuration for our grounded proposer.")
                    # Randomly select the tip
                    selected_tip_key = self.rng.choice(list(TIPS.keys()))
                    selected_tip = TIPS[selected_tip_key]
                    self.use_tip = bool(
                        selected_tip,
                    )
                    if self.verbose: print(f"Selected tip: {selected_tip_key}")

                proposed_instructions[pred_i].append(
                    self.propose_instruction_for_predictor(
                        program=program,
                        predictor=predictor,
                        pred_i=pred_i,
                        # prompt_model=prompt_model,
                        T=T,
                        demo_candidates=demo_candidates,
                        demo_set_i=demo_set_i,
                        trial_logs=trial_logs,
                        tip=selected_tip,
                    ),
                )

        return proposed_instructions

    def propose_instruction_for_predictor(
        self,
        program,
        predictor,
        pred_i,
        # prompt_model,
        T,
        demo_candidates,
        demo_set_i,
        trial_logs,
        tip=None,
    ):
        """This method is responsible for returning a single instruction for a given predictor, using the specified criteria."""

        # Create an instruction history string for our predictor
        instruction_history = create_predictor_level_history_string(
            program, pred_i, trial_logs, MAX_INSTRUCT_IN_HISTORY,
        )

        # Create our instruction generator class (given specific criteria for this round of proposal)
        instruction_generator = GenerateModuleInstruction(
            program_code_string=self.program_code_string,
            use_dataset_summary=self.use_dataset_summary,
            program_aware=self.program_aware,
            use_task_demos=self.use_task_demos,
            use_instruct_history=self.use_instruct_history and instruction_history,
            use_tip=self.use_tip,
            verbose=self.verbose
        )

        # Generate a new instruction for our predictor, using the temperature specified for this round
        original_temp = self.prompt_model.kwargs["temperature"]

        with dspy.settings.context(lm=self.prompt_model):
            self.prompt_model.kwargs["temperature"] = T
            proposed_instruction = instruction_generator.forward(
                demo_candidates=demo_candidates,
                pred_i=pred_i,
                demo_set_i=demo_set_i,
                program=program,
                data_summary=self.data_summary,
                previous_instructions=instruction_history,
                tip=tip,
            ).proposed_instruction
        self.prompt_model.kwargs["temperature"] = original_temp

        # Log the trace used to generate the new instruction, along with the new instruction itself
        if self.verbose: self.prompt_model.inspect_history(n=1)
        if self.verbose: print(f"PROPOSED INSTRUCTION: {proposed_instruction}")

        return strip_prefix(proposed_instruction)



import json
from typing import Optional, Union

import requests
from requests.auth import HTTPBasicAuth

import dspy
from dsp.utils import dotdict


class WatsonDiscoveryRM(dspy.Retrieve):
    """A retrieval module that uses Watson Discovery to return the top passages for a given query.

    Args:
        apikey (str): apikey for authentication purposes,
        url (str): endpoint URL that includes the service instance ID
        version (str): Release date of the version of the API you want to use. Specify dates in YYYY-MM-DD format.
        project_id (str): The Universally Unique Identifier (UUID) of the project.
        collection_ids (list): An array containing the collections on which the search will be executed.
        k (int, optional): The number of top passages to retrieve. Defaults to 5.

    Returns:
        dspy.Prediction: An object containing the retrieved passages.
    """

    def __init__(
        self,
        apikey: str,
        url:str,
        version:str,
        project_id:str,
        collection_ids:list=None,
        k: int = 5,
    ):
        if collection_ids is None:
            collection_ids = []
        self.apikey=apikey
        self.url=url,
        self.version:version
        self.project_id:project_id
        self.collection_ids=collection_ids
        self.k: k
        self.query_url=url + "/v2/projects/" + project_id + "/query?version=" + version
        
        super().__init__(k=k)

    def forward(self, query_or_queries: Union[str, list[str]], k: Optional[int]= None) -> dspy.Prediction:
        """Search with Watson Discovery for self.k top passages for query.

        Args:
            query_or_queries (Union[str, List[str]]): The query or queries to search for.
            k (int, optional): The number of top passages to retrieve.

        Returns:
            dspy.Prediction: An object containing the retrieved passages.
        """
        queries = [query_or_queries] if isinstance(query_or_queries, str) else query_or_queries
        queries = [q for q in queries if q]  # Filter empty queries
        k = self.k if k is None else k

        response=[]
        try:
            for query in queries:
                payload = json.dumps({
                    "collection_ids": self.collection_ids,
                    "natural_language_query": "text:"+query,
                    "passages": {
                        "enabled": True,
                        "count": k,
                        "fields":["text"],
                        "characters": 500,
                        "find_answers": False,
                        "per_document": False,
                    },
                    "highlight": False,
                })
                headers = {'Content-Type': 'application/json'}

                discovery_results = requests.request(
                    "POST",
                    url=self.query_url, 
                    headers=headers,
                    auth=HTTPBasicAuth("apikey", self.apikey),
                    data=payload,
                )
                
                discovery_results.raise_for_status()

                doc_dict={}
                for d in discovery_results.json()["results"]:
                    doc_dict[d["document_id"]]=d["title"]

                for d in discovery_results.json()["passages"]:
                    response.append(dotdict({
                        "title":doc_dict[d["document_id"]],
                        "long_text": doc_dict[d["document_id"]]+" | " + d["passage_text"],
                        "passage_score": d["passage_score"],
                        "document_id": d["document_id"],
                        "collection_id": d["collection_id"],
                        "start_offset": d["start_offset"],
                        "end_offset": d["end_offset"],
                        "field": d["field"],
                    }))

        except requests.exceptions.RequestException as err:
            raise SystemExit(err) from err
        return response


from typing import Any, Optional

import backoff

from dsp.modules.lm import LM
from dsp.utils.settings import settings

try:
    import cohere
    cohere_api_error = cohere.errors.UnauthorizedError
except ImportError:
    cohere_api_error = Exception
    # print("Not loading Cohere because it is not installed.")
except AttributeError:
    cohere_api_error = Exception


def backoff_hdlr(details):
    """Handler from https://pypi.org/project/backoff/"""
    print(
        "Backing off {wait:0.1f} seconds after {tries} tries "
        "calling function {target} with kwargs "
        "{kwargs}".format(**details),
    )


def giveup_hdlr(details):
    """wrapper function that decides when to give up on retry"""
    if "rate limits" in details.message:
        return False
    return True


class Cohere(LM):
    """Wrapper around Cohere's API.

    Currently supported models include `command-r-plus`, `command-r`, `command`, `command-nightly`, `command-light`, `command-light-nightly`.
    """

    def __init__(
        self,
        model: str = "command-r",
        api_key: Optional[str] = None,
        stop_sequences: list[str] = [],
        **kwargs,
    ):
        """
        Parameters
        ----------
        model : str
            Which pre-trained model from Cohere to use?
            Choices are [`command-r-plus`, `command-r`, `command`, `command-nightly`, `command-light`, `command-light-nightly`]
        api_key : str
            The API key for Cohere.
            It can be obtained from https://dashboard.cohere.ai/register.
        stop_sequences : list of str
            Additional stop tokens to end generation.
        **kwargs: dict
            Additional arguments to pass to the API provider.
        """
        super().__init__(model)
        self.co = cohere.Client(api_key, client_name='dspy')
        self.provider = "cohere"
        self.kwargs = {
            "model": model,
            "temperature": 0.0,
            "max_tokens": 2000,
            "p": 1,
            "num_generations": 1,
            **kwargs,
        }
        self.stop_sequences = stop_sequences
        self.max_num_generations = 5

        self.history: list[dict[str, Any]] = []

    def basic_request(self, prompt: str, **kwargs):
        raw_kwargs = kwargs
        kwargs = {
            **self.kwargs,
            "stop_sequences": self.stop_sequences,
            "chat_history": [],
            "message": prompt,
            **kwargs,
        }
        kwargs.pop("num_generations")
        if "n" in kwargs.keys():
            kwargs.pop("n")
        if "stop" in kwargs.keys():
            kwargs.pop("stop")
        response = self.co.chat(**kwargs)

        self.history.append(
            {
                "prompt": prompt,
                "response": response,
                "kwargs": kwargs,
                "raw_kwargs": raw_kwargs,
            },
        )

        return response

    @backoff.on_exception(
        backoff.expo,
        (cohere_api_error),
        max_time=settings.backoff_time,
        on_backoff=backoff_hdlr,
        giveup=giveup_hdlr,
    )
    def request(self, prompt: str, **kwargs):
        """Handles retrieval of completions from Cohere whilst handling API errors"""
        return self.basic_request(prompt, **kwargs)

    def __call__(
        self,
        prompt: str,
        **kwargs,
    ):
        response = self.request(prompt, **kwargs)
        return [response.text]



import json
import re
from typing import Any, Optional, Type, Union

from pydantic import BaseModel, Field, ValidationError

import dspy
from dsp.utils import dotdict

try:
    from snowflake.core import Root
    from snowflake.snowpark.functions import col
except ImportError:
    raise ImportError(
        "The snowflake-snowpark-python library is required to use SnowflakeRM. Install it with dspy-ai[snowflake]",
    )


class SnowflakeRM(dspy.Retrieve):
    """A retrieval module that uses Snowflake's Cortex Search service to return the top relevant passages for a given query.

    Assumes that a Snowflake Cortex Search endpoint has been configured by the use.

    For more information on configuring the Cortex Search service, visit: https://docs.snowflake.com/en/user-guide/snowflake-cortex/cortex-search/cortex-search-overview

    Args:
        snowflake_session (object): Snowflake Snowpark session for accessing the service.
        cortex_search_service(str): Name of the Cortex Search service to be used.
        snowflake_database (str): The name of the Snowflake table containing document embeddings.
        snowflake_schema (str): The name of the Snowflake table containing document embeddings.
        auto_filter (bool): Auto generate metadata filter based on user query and push it down prior to retrieving Cortex Search results.
        k (int, optional): The default number of top passages to retrieve. Defaults to 3.
    """

    def __init__(
        self,
        snowflake_session: object,
        cortex_search_service: str,
        snowflake_database: str,
        snowflake_schema: str,
        auto_filter=False,
        k: int = 3,
        max_retries=3,
    ):
        super().__init__(k=k)
        self.k = k
        self.auto_filter = auto_filter
        self.max_retries = max_retries
        self.cortex_search_service_name = cortex_search_service
        self.client = self._fetch_cortex_service(
            snowflake_session, snowflake_database, snowflake_schema, cortex_search_service
        )

        if self.auto_filter:
            self.sample_values = self._get_sample_values(
                snowpark_session=snowflake_session, cortex_search_service=cortex_search_service
            )
            self.optimized_filter_gen = SmartSearch()
            self.filter_lm = dspy.Snowflake(session=snowflake_session, model="mixtral-8x7b")

    def forward(
        self,
        query_or_queries: Union[str, list[str]],
        retrieval_columns: list[str],
        filter: Optional[dict] = None,
        k: Optional[int] = None,
    ) -> dspy.Prediction:
        """Query Cortex Search endpoint for top k relevant passages.
        Args:
            query_or_queries (Union[str, List[str]]): The query or queries to search for.
            retrieval_columns (List[str]): Columns to include in response.
            filter (Optional[json]):Filter query.
            k (Optional[int]): The number of top passages to retrieve. Defaults to self.k.
        Returns:
            dspy.Prediction: An object containing the retrieved passages.
        """
        k = self.k if k is None else k
        queries = [query_or_queries] if isinstance(query_or_queries, str) else query_or_queries
        queries = [q for q in queries if q]
        passages = []

        for cortex_query in queries:
            if self.auto_filter:
                response_chunks = self._cortex_search_with_auto_filter(
                    cortex_search_service=self.client, query=cortex_query, columns=retrieval_columns, k=k
                )

            else:
                response_chunks = self._query_cortex_search(
                    cortex_search_service=self.client,
                    query=cortex_query,
                    columns=retrieval_columns,
                    filter=filter,
                    k=k,
                )

            if len(retrieval_columns) == 1:
                passages.extend(
                    dotdict({"long_text": passage[retrieval_columns[0]]}) for passage in response_chunks["results"]
                )
            else:
                passages.extend(dotdict({"long_text": str(passage)}) for passage in response_chunks["results"])

        return passages

    def _fetch_cortex_service(self, snowpark_session, snowflake_database, snowflake_schema, cortex_search_service_name):
        """Fetch the Cortex Search service to be used"""
        snowpark_session.query_tag = {"origin": "sf_sit", "name": "dspy", "version": {"major": 1, "minor": 0}}
        root = Root(snowpark_session)

        # fetch service
        search_service = (
            root.databases[snowflake_database]
            .schemas[snowflake_schema]
            .cortex_search_services[cortex_search_service_name]
        )

        return search_service

    def _query_cortex_search(self, cortex_search_service, query, columns, filter, k):
        """Query Cortex Search endpoint for top relevant passages .

        Args:
            cortex_search_service (object): cortex search service for querying
            query (str): The query or queries to search for.
            columns (Optional[list]): A comma-separated list of columns to return for each relevant result in the response. These columns must be included in the source query for the service.
            filter (Optional[json]): A filter object for filtering results based on data in the ATTRIBUTES columns. See Filter syntax.
            k (Optional[int]): The number of top passages to retrieve. Defaults to self.k.

        Returns:
            dspy.Prediction: An object containing the retrieved passages.
        """
        # query service
        resp = cortex_search_service.search(query=query, columns=columns, filter=filter, limit=k)

        return json.loads(resp.to_json())

    def _cortex_search_with_auto_filter(self, cortex_search_service, query, columns, k):
        """Cortex Search Query with automatic metadata filter generation."""

        search_response = None
        with dspy.settings.context(lm=self.filter_lm):
            if self.auto_filter:
                for _ in range(self.max_retries):
                    raw_filter_query = self.optimized_filter_gen(
                        query=query, attributes=str([*self.sample_values]), sample_values=str(self.sample_values)
                    )["answer"]

                    try:
                        filter_query = json.loads(raw_filter_query)
                        search_response = self._query_cortex_search(
                            cortex_search_service=cortex_search_service,
                            query=query,
                            columns=columns,
                            filter=filter_query,
                            k=k,
                        )
                        break
                    except json.JSONDecodeError:
                        continue
                    except Exception:
                        continue

            else:
                search_response = self._query_cortex_search(
                    cortex_search_service=cortex_search_service, query=query, columns=columns, filter=None, k=k
                )

        return search_response

    def _get_search_attributes(self, snowpark_session, search_service_name):
        df = snowpark_session.sql("SHOW CORTEX SEARCH SERVICES")
        raw_atts = (
            df.where(col('"name"') == search_service_name).select('"attribute_columns"').to_pandas().loc[0].values[0]
        )
        attribute_list = raw_atts.split(",")

        return attribute_list

    def _get_search_table(self, snowpark_session, search_service_name):
        df = snowpark_session.sql("SHOW CORTEX SEARCH SERVICES")
        table_def = df.where(col('"name"') == search_service_name).select('"definition"').to_pandas().loc[0].values[0]

        pattern = r"FROM\s+([\w\.]+)"
        match = re.search(pattern, table_def)

        if match:
            from_value = match.group(1)
            return from_value
        else:
            print("No match found.")

        return table_def

    def _get_sample_values(self, snowpark_session, cortex_search_service, max_samples=10):
        sample_values = {}
        attributes = self._get_search_attributes(
            snowpark_session=snowpark_session, search_service_name=cortex_search_service
        )
        table_name = self._get_search_table(
            snowpark_session=snowpark_session, search_service_name=cortex_search_service
        )

        for attribute in attributes:
            query = f"""SELECT DISTINCT({attribute}) FROM {table_name} LIMIT {max_samples}"""
            sample_values[attribute] = list(snowpark_session.sql(query).to_pandas()[attribute].values)

        return sample_values


class JSONFilter(BaseModel):
    answer: str = Field(description="The filter_query in valid JSON format")

    @classmethod
    def model_validate_json(cls, json_data: str, *, strict: bool | None = None, context: dict[str, Any] | None = None):
        __tracebackhide__ = True
        try:
            return cls.__pydantic_validator__.validate_json(json_data, strict=strict, context=context)
        except ValidationError:
            min_length = get_min_length(cls)
            for substring_length in range(len(json_data), min_length - 1, -1):
                for start in range(len(json_data) - substring_length + 1):
                    substring = json_data[start : start + substring_length]
                    try:
                        res = cls.__pydantic_validator__.validate_json(substring, strict=strict, context=context)
                        return res
                    except ValidationError:
                        pass
        raise ValueError("Could not find valid json")


class GenerateFilter(dspy.Signature):
    """
    Given a query, attributes in the data, and example values of each attribute, generate a filter in valid JSON format.
    Ensure the filter only uses valid operators: @eq, @contains,@and,@or,@not
    Ensure only the valid JSON is output with no other reasoning.

    ---
    Query: What was the sentiment of CEOs between 2021 and 2024?
    Attributes: industry,hq,date
    Sample Values: {"industry":["biotechnology","healthcare","agriculture"],"HQ":["NY, US","CA,US","FL,US"],"date":["01/01,1999","01/01/2024"]}
    Answer: {"@or":[{"@eq":{"year":"2021"}},{"@eq":{"year":"2022"}},{"@eq":{"year":"2023"}},{"@eq":{"year":"2024"}}]}

    Query: What is the sentiment of Biotech CEO's of companies based in New York?
    Attributes: industry,hq,date
    Sample Values: {"industry":["biotechnology","healthcare","agriculture"],"HQ":["NY, US","CA,US","FL,US"],"date":["01/01,1999","01/01/2024"]}
    Answer: {"@and": [ { "@eq": { "industry"": "biotechnology" } }, { "@eq": { "HQ": "NY,US" } }]}

    Query: What is the sentiment of Biotech CEOs outside of California?
    Attributes: industry,hq,date
    Sample Values: {"industry":["biotechnology","healthcare","agriculture"],"HQ":["NY, US","CA,US","FL,US"],"date":["01/01,1999","01/01/2024"]}
    Answer: {"@and":[{ "@eq": { "industry": "biotechnology" } },{"@not":{"@eq":{"HQ":"CA,US"}}}]}

    Query: What is the sentiment of Biotech CEOs outside of California?
    Attributes: industry,hq,date
    Sample Values: {"industry":["biotechnology","healthcare","agriculture"],"HQ":["NY, US","CA,US","FL,US"],"date":["01/01,1999","01/01/2024"]}
    Answer: {"@and":[{ "@eq": { "industry": "biotechnology" } },{"@not":{"@eq":{"HQ":"CA,US"}}}]}

    Query: What is sentiment towards ag and biotech companies based outside of the US?
    Attributes: industry,hq,date
    Sample Values: {"industry"":["biotechnology","healthcare","agriculture"],"COUNTRY":["United States","Ireland","Russia","Georgia","Spain"],"month":["01","02","03","06","11","12""],""year"":["2022","2023","2024"]}
    Answer:{"@and": [{ "@or": [{"@eq":{ "industry": "biotechnology" } },{"@eq":{"industry":"agriculture"}}]},{ "@not": {"@eq": { "COUNTRY": "United States" } }}]}

    """

    query = dspy.InputField(desc="user query")
    attributes = dspy.InputField(desc="attributes to filter on")
    sample_values = dspy.InputField(desc="examples of values per attribute")
    answer: JSONFilter = dspy.OutputField(
        desc="filter query in valid JSON format. ONLY output the filter query in JSON, no reasoning"
    )


class SmartSearch(dspy.Module):
    def __init__(self):
        super().__init__()
        self.filter_gen = dspy.ChainOfThought(GenerateFilter)

    def forward(self, query, attributes, sample_values):
        filter_query = self.filter_gen(query=query, attributes=attributes, sample_values=sample_values)

        return filter_query


def get_min_length(model: Type[BaseModel]):
    min_length = 0
    for key, field in model.model_fields.items():
        if issubclass(field.annotation, BaseModel):
            min_length += get_min_length(field.annotation)
        min_length += len(key)
    return min_length



import logging
from typing import Any

import backoff

try:
    import groq
    from groq import Groq

    groq_api_error = (groq.APIError, groq.RateLimitError)
except ImportError:
    groq_api_error = Exception


from dsp.modules.lm import LM
from dsp.utils.settings import settings


def backoff_hdlr(details):
    """Handler from https://pypi.org/project/backoff/"""
    print(
        "Backing off {wait:0.1f} seconds after {tries} tries "
        "calling function {target} with kwargs "
        "{kwargs}".format(**details),
    )


class GroqLM(LM):
    """Wrapper around groq's API.

    Args:
        model (str, optional): groq supported LLM model to use. Defaults to "mixtral-8x7b-32768".
        api_key (Optional[str], optional): API provider Authentication token. use Defaults to None.
        **kwargs: Additional arguments to pass to the API provider.
    """

    def __init__(
        self,
        api_key: str,
        model: str = "mixtral-8x7b-32768",
        **kwargs,
    ):
        super().__init__(model)
        self.provider = "groq"
        if api_key:
            self.api_key = api_key
            self.client = Groq(api_key=api_key)
        else:
            raise ValueError("api_key is required for groq")

        self.kwargs = {
            "temperature": 0.0,
            "max_tokens": 150,
            "top_p": 1,
            "frequency_penalty": 0,
            "presence_penalty": 0,
            "n": 1,
            **kwargs,
        }
        models = self.client.models.list().data
        if models is not None:
            if model in [m.id for m in models]:
                self.kwargs["model"] = model
        self.history: list[dict[str, Any]] = []

    def log_usage(self, response):
        """Log the total tokens from the Groq API response."""
        usage_data = response.usage  # Directly accessing the 'usage' attribute
        if usage_data:
            total_tokens = usage_data.total_tokens
            logging.debug(f"Groq Total Tokens Response Usage: {total_tokens}")

    def basic_request(self, prompt: str, **kwargs):
        raw_kwargs = kwargs

        kwargs = {**self.kwargs, **kwargs}

        kwargs["messages"] = [{"role": "user", "content": prompt}]
        response = self.chat_request(**kwargs)

        history = {
            "prompt": prompt,
            "response": response.choices[0].message.content,
            "kwargs": kwargs,
            "raw_kwargs": raw_kwargs,
        }

        self.history.append(history)

        return response

    @backoff.on_exception(
        backoff.expo,
        groq_api_error,
        max_time=settings.backoff_time,
        on_backoff=backoff_hdlr,
    )
    def request(self, prompt: str, **kwargs):
        """Handles retrieval of model completions whilst handling rate limiting and caching."""
        if "model_type" in kwargs:
            del kwargs["model_type"]

        return self.basic_request(prompt, **kwargs)

    def _get_choice_text(self, choice) -> str:
        return choice.message.content

    def chat_request(self, **kwargs):
        """Handles retrieval of model completions whilst handling rate limiting and caching."""
        response = self.client.chat.completions.create(**kwargs)
        return response

    def __call__(
        self,
        prompt: str,
        only_completed: bool = True,
        return_sorted: bool = False,
        **kwargs,
    ) -> list[dict[str, Any]]:
        """Retrieves completions from model.

        Args:
            prompt (str): prompt to send to model
            only_completed (bool, optional): return only completed responses and ignores completion due to length. Defaults to True.
            return_sorted (bool, optional): sort the completion choices using the returned probabilities. Defaults to False.

        Returns:
            list[dict[str, Any]]: list of completion choices
        """

        assert only_completed, "for now"
        assert return_sorted is False, "for now"
        response = self.request(prompt, **kwargs)

        self.log_usage(response)

        choices = response.choices

        completions = [self._get_choice_text(c) for c in choices]
        if return_sorted and kwargs.get("n", 1) > 1:
            scored_completions = []

            for c in choices:
                tokens, logprobs = (
                    c["logprobs"]["tokens"],
                    c["logprobs"]["token_logprobs"],
                )

                if "<|endoftext|>" in tokens:
                    index = tokens.index("<|endoftext|>") + 1
                    tokens, logprobs = tokens[:index], logprobs[:index]

                avglog = sum(logprobs) / len(logprobs)
                scored_completions.append((avglog, self._get_choice_text(c)))

            scored_completions = sorted(scored_completions, reverse=True)
            completions = [c for _, c in scored_completions]

        return completions



from abc import ABC, abstractmethod

from dspy.utils.callback import with_callbacks

class Adapter(ABC):
    def __init__(self, callbacks=None):
        self.callbacks = callbacks or []

    def __init_subclass__(cls, **kwargs) -> None:
        super().__init_subclass__(**kwargs)

        # Decorate format() and parse() method with with_callbacks
        cls.format = with_callbacks(cls.format)
        cls.parse = with_callbacks(cls.parse)

    def __call__(self, lm, lm_kwargs, signature, demos, inputs, _parse_values=True):
        inputs_ = self.format(signature, demos, inputs)
        inputs_ = dict(prompt=inputs_) if isinstance(inputs_, str) else dict(messages=inputs_)

        outputs = lm(**inputs_, **lm_kwargs)
        values = []

        try:
            for output in outputs:
                value = self.parse(signature, output, _parse_values=_parse_values)
                assert set(value.keys()) == set(signature.output_fields.keys()), f"Expected {signature.output_fields.keys()} but got {value.keys()}"
                values.append(value)
            return values

        except Exception as e:
            from .json_adapter import JSONAdapter
            if _parse_values and not isinstance(self, JSONAdapter):
                return JSONAdapter()(lm, lm_kwargs, signature, demos, inputs, _parse_values=_parse_values)
            raise e

    @abstractmethod
    def format(self, signature, demos, inputs):
       raise NotImplementedError

    @abstractmethod
    def parse(self, signature, completion, _parse_values):
       raise NotImplementedError

    def format_finetune_data(self, signature, demos, inputs, outputs):
        raise NotImplementedError



import functools
import logging
import os
import threading
import uuid
from datetime import datetime
from typing import Any, Dict, List, Literal, Optional

import litellm
import ujson

from dspy.adapters.base import Adapter
from dspy.clients.openai import OpenAIProvider
from dspy.clients.provider import Provider, TrainingJob
from dspy.clients.utils_finetune import DataFormat, infer_data_format, validate_data_format
from dspy.utils.callback import BaseCallback, with_callbacks

from .base_lm import BaseLM

logger = logging.getLogger(__name__)


class LM(BaseLM):
    """
    A language model supporting chat or text completion requests for use with DSPy modules.
    """

    def __init__(
        self,
        model: str,
        model_type: Literal["chat", "text"] = "chat",
        temperature: float = 0.0,
        max_tokens: int = 1000,
        cache: bool = True,
        callbacks: Optional[List[BaseCallback]] = None,
        num_retries: int = 3,
        provider=None,
        finetuning_model: Optional[str] = None,
        launch_kwargs: Optional[dict[str, Any]] = None,
        **kwargs,
    ):
        """
        Create a new language model instance for use with DSPy modules and programs.

        Args:
            model: The model to use. This should be a string of the form ``"llm_provider/llm_name"``
                   supported by LiteLLM. For example, ``"openai/gpt-4o"``.
            model_type: The type of the model, either ``"chat"`` or ``"text"``.
            temperature: The sampling temperature to use when generating responses.
            max_tokens: The maximum number of tokens to generate per response.
            cache: Whether to cache the model responses for reuse to improve performance
                   and reduce costs.
            callbacks: A list of callback functions to run before and after each request.
            num_retries: The number of times to retry a request if it fails transiently due to
                         network error, rate limiting, etc. Requests are retried with exponential
                         backoff.
            provider: The provider to use. If not specified, the provider will be inferred from the model.
            finetuning_model: The model to finetune. In some providers, the models available for finetuning is different
                from the models available for inference.
        """
        # Remember to update LM.copy() if you modify the constructor!
        self.model = model
        self.model_type = model_type
        self.cache = cache
        self.provider = provider or self.infer_provider()
        self.callbacks = callbacks or []
        self.kwargs = dict(temperature=temperature, max_tokens=max_tokens, **kwargs)
        self.history = []
        self.callbacks = callbacks or []
        self.num_retries = num_retries
        self.finetuning_model = finetuning_model
        self.launch_kwargs = launch_kwargs

        # TODO(bug): Arbitrary model strings could include the substring "o1-".
        # We should find a more robust way to check for the "o1-" family models.
        if "o1-" in model:
            assert (
                max_tokens >= 5000 and temperature == 1.0
            ), "OpenAI's o1-* models require passing temperature=1.0 and max_tokens >= 5000 to `dspy.LM(...)`"

    @with_callbacks
    def __call__(self, prompt=None, messages=None, **kwargs):
        # Build the request.
        cache = kwargs.pop("cache", self.cache)
        messages = messages or [{"role": "user", "content": prompt}]
        kwargs = {**self.kwargs, **kwargs}

        # Make the request and handle LRU & disk caching.
        if self.model_type == "chat":
            completion = cached_litellm_completion if cache else litellm_completion
        else:
            completion = cached_litellm_text_completion if cache else litellm_text_completion

        response = completion(
            request=ujson.dumps(dict(model=self.model, messages=messages, **kwargs)),
            num_retries=self.num_retries,
        )
        outputs = [c.message.content if hasattr(c, "message") else c["text"] for c in response["choices"]]

        # Logging, with removed api key & where `cost` is None on cache hit.
        kwargs = {k: v for k, v in kwargs.items() if not k.startswith("api_")}
        entry = dict(prompt=prompt, messages=messages, kwargs=kwargs, response=response)
        entry = dict(**entry, outputs=outputs, usage=dict(response["usage"]))
        entry = dict(**entry, cost=response.get("_hidden_params", {}).get("response_cost"))
        entry = dict(
            **entry,
            timestamp=datetime.now().isoformat(),
            uuid=str(uuid.uuid4()),
            model=self.model,
            model_type=self.model_type,
        )
        self.history.append(entry)
        self.update_global_history(entry)

        return outputs

    def launch(self, launch_kwargs: Optional[Dict[str, Any]] = None):
        launch_kwargs = launch_kwargs or self.launch_kwargs
        self.provider.launch(self.model, launch_kwargs)

    def kill(self, launch_kwargs: Optional[Dict[str, Any]] = None):
        launch_kwargs = launch_kwargs or self.launch_kwargs
        self.provider.kill(self.model, launch_kwargs)

    def finetune(
        self,
        train_data: List[Dict[str, Any]],
        train_kwargs: Optional[Dict[str, Any]] = None,
        data_format: Optional[DataFormat] = None,
    ) -> TrainingJob:
        from dspy import settings as settings

        err = "Fine-tuning is an experimental feature."
        err += " Set `dspy.settings.experimental` to `True` to use it."
        assert settings.experimental, err

        err = f"Provider {self.provider} does not support fine-tuning."
        assert self.provider.finetunable, err

        # Perform data validation before starting the thread to fail early
        train_kwargs = train_kwargs or {}
        if not data_format:
            adapter = self.infer_adapter()
            data_format = infer_data_format(adapter)
        validate_data_format(data=train_data, data_format=data_format)

        # TODO(PR): We can quickly add caching, but doing so requires
        # adding functions that just call other functions as we had in the last
        # iteration, unless people have other ideas.
        def thread_function_wrapper():
            return self._run_finetune_job(job)

        thread = threading.Thread(target=thread_function_wrapper)
        model_to_finetune = self.finetuning_model or self.model
        job = self.provider.TrainingJob(
            thread=thread, model=model_to_finetune, train_data=train_data, train_kwargs=train_kwargs, data_format=data_format
        )
        thread.start()

        return job

    def _run_finetune_job(self, job: TrainingJob):
        # TODO(enhance): We should listen for keyboard interrupts somewhere.
        # Requires TrainingJob.cancel() to be implemented for each provider.
        try:
            model = self.provider.finetune(
                job=job,
                model=job.model,
                train_data=job.train_data,
                train_kwargs=job.train_kwargs,
                data_format=job.data_format,
            )
            lm = self.copy(model=model)
            job.set_result(lm)
        except Exception as err:
            logger.error(err)
            job.set_result(err)

    def infer_provider(self) -> Provider:
        if OpenAIProvider.is_provider_model(self.model):
            return OpenAIProvider()
        # TODO(PR): Keeping this function here will require us to import all
        # providers in this file. Is this okay?
        return Provider()

    def infer_adapter(self) -> Adapter:
        import dspy

        if dspy.settings.adapter:
            return dspy.settings.adapter

        model_type_to_adapter = {
            "chat": dspy.ChatAdapter(),
        }
        model_type = self.model_type
        return model_type_to_adapter[model_type]

    def copy(self, **kwargs):
        """Returns a copy of the language model with possibly updated parameters."""

        import copy

        new_instance = copy.deepcopy(self)
        new_instance.history = []

        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(new_instance, key, value)
            if (key in self.kwargs) or (not hasattr(self, key)):
                new_instance.kwargs[key] = value

        return new_instance


@functools.lru_cache(maxsize=None)
def cached_litellm_completion(request, num_retries: int):
    return litellm_completion(
        request,
        cache={"no-cache": False, "no-store": False},
        num_retries=num_retries,
    )


def litellm_completion(request, num_retries: int, cache={"no-cache": True, "no-store": True}):
    kwargs = ujson.loads(request)
    return litellm.completion(
        num_retries=num_retries,
        cache=cache,
        **kwargs,
    )


@functools.lru_cache(maxsize=None)
def cached_litellm_text_completion(request, num_retries: int):
    return litellm_text_completion(
        request,
        num_retries=num_retries,
        cache={"no-cache": False, "no-store": False},
    )


def litellm_text_completion(request, num_retries: int, cache={"no-cache": True, "no-store": True}):
    kwargs = ujson.loads(request)

    # Extract the provider and model from the model string.
    # TODO: Not all the models are in the format of "provider/model"
    model = kwargs.pop("model").split("/", 1)
    provider, model = model[0] if len(model) > 1 else "openai", model[-1]

    # Use the API key and base from the kwargs, or from the environment.
    api_key = kwargs.pop("api_key", None) or os.getenv(f"{provider}_API_KEY")
    api_base = kwargs.pop("api_base", None) or os.getenv(f"{provider}_API_BASE")

    # Build the prompt from the messages.
    prompt = "\n\n".join([x["content"] for x in kwargs.pop("messages")] + ["BEGIN RESPONSE:"])

    return litellm.text_completion(
        cache=cache,
        model=f"text-completion-openai/{model}",
        api_key=api_key,
        api_base=api_base,
        prompt=prompt,
        num_retries=num_retries,
        **kwargs,
    )



import threading
from contextlib import contextmanager
from copy import deepcopy

from dsp.utils.utils import dotdict

DEFAULT_CONFIG = dotdict(
    lm=None,
    adapter=None,
    rm=None,
    branch_idx=0,
    reranker=None,
    compiled_lm=None,
    force_reuse_cached_compilation=False,
    compiling=False,
    skip_logprobs=False,
    trace=[],
    release=0,
    bypass_assert=False,
    bypass_suggest=False,
    assert_failures=0,
    suggest_failures=0,
    langchain_history=[],
    experimental=False,
    backoff_time=10,
    callbacks=[],
    async_max_workers=8,
)


class Settings:
    """DSP configuration settings."""

    _instance = None

    def __new__(cls):
        """
        Singleton Pattern. See https://python-patterns.guide/gang-of-four/singleton/
        """

        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance.lock = threading.Lock()
            cls._instance.main_tid = threading.get_ident()
            cls._instance.main_stack = []
            cls._instance.stack_by_thread = {}
            cls._instance.stack_by_thread[threading.get_ident()] = cls._instance.main_stack

            #  TODO: remove first-class support for re-ranker and potentially combine with RM to form a pipeline of sorts
            #  eg: RetrieveThenRerankPipeline(RetrievalModel, Reranker)
            #  downstream operations like dsp.retrieve would use configs from the defined pipeline.

            # make a deepcopy of the default config to avoid modifying the default config
            cls._instance.__append(deepcopy(DEFAULT_CONFIG))

        return cls._instance

    @property
    def config(self):
        thread_id = threading.get_ident()
        if thread_id not in self.stack_by_thread:
            self.stack_by_thread[thread_id] = [self.main_stack[-1].copy()]
        return self.stack_by_thread[thread_id][-1]

    def __getattr__(self, name):
        if hasattr(self.config, name):
            return getattr(self.config, name)

        if name in self.config:
            return self.config[name]

        super().__getattr__(name)

    def __append(self, config):
        thread_id = threading.get_ident()
        if thread_id not in self.stack_by_thread:
            self.stack_by_thread[thread_id] = [self.main_stack[-1].copy()]
        self.stack_by_thread[thread_id].append(config)

    def __pop(self):
        thread_id = threading.get_ident()
        if thread_id in self.stack_by_thread:
            self.stack_by_thread[thread_id].pop()

    def configure(self, inherit_config: bool = True, **kwargs):
        """Set configuration settings.

        Args:
            inherit_config (bool, optional): Set configurations for the given, and use existing configurations for the rest. Defaults to True.
        """
        if inherit_config:
            config = {**self.config, **kwargs}
        else:
            config = {**kwargs}

        self.__append(config)

    @contextmanager
    def context(self, inherit_config=True, **kwargs):
        self.configure(inherit_config=inherit_config, **kwargs)

        try:
            yield
        finally:
            self.__pop()

    def __repr__(self) -> str:
        return repr(self.config)


settings = Settings()


from .avatar_optimizer import *
from .bettertogether import BetterTogether
from .bootstrap import *
from .bootstrap_finetune import BootstrapFinetune
from .copro_optimizer import COPRO
from .ensemble import *
from .knn_fewshot import *
from .mipro_optimizer import MIPRO
from .mipro_optimizer_v2 import MIPROv2
from .random_search import *
from .signature_opt import SignatureOptimizer
from .signature_opt_bayesian import BayesianSignatureOptimizer
from .teleprompt import *
from .teleprompt_optuna import *
from .vanilla import *
